from annet.annlib.command import Command, CommandList
from annet.annlib.netdev.views.hardware import HardwareView
from annet.vendors.tabparser import CiscoFormatter
from annet.vendors.base import AbstractVendor
from annet.vendors.registry import registry


@registry.register
class CiscoVendor(AbstractVendor):
    NAME = "cisco"

    def apply(self, hw: HardwareView, do_commit: bool, do_finalize: bool, path: str) -> tuple[CommandList, CommandList]:
        before, after = CommandList(), CommandList()

        before.add_cmd(Command("conf t"))
        after.add_cmd(Command("exit"))
        if do_finalize:
            after.add_cmd(Command("copy running-config startup-config", timeout=40))

        return before, after

    def match(self) -> list[str]:
        return ["Cisco"]

    @property
    def reverse(self) -> str:
        return "no"

    @property
    def hardware(self) -> HardwareView:
        return HardwareView("Cisco")

    def svi_name(self, num: int) -> str:
        return f"Vlan{num}"

    def make_formatter(self, **kwargs) -> CiscoFormatter:
        return CiscoFormatter(**kwargs)

    @property
    def exit(self) -> str:
        return "exit"



import functools
import re
from collections import OrderedDict as odict

from annet.annlib import lib
from annet.vendors import tabparser


# =====
def parse_text(text, params_scheme):
    return _parse_tree_with_params(tabparser.parse_to_tree(text, _split_rows, ["#"]), params_scheme)


@functools.lru_cache()
def compile_row_regexp(row, flags=0):
    if "(?i)" in row:
        row = row.replace("(?i)", "")
        flags |= re.IGNORECASE

    if "*" in row:
        row = re.sub(r"\(([^\?])", r"(?:\1", row)  # Все дефолтные группы превратить в non-captured
        row = re.sub(r"\*/(\S+)/", r"(\1)", row)   # */{regex_one_word}/ -> ({regex_one_word})
        row = re.sub(r"(^|\s)\*", r"\1([^\\s]+)", row)

    # Заменяем <someting> на named-группы
    row = re.sub(r"<(\w+)>", r"(?P<\1>\\w+)", row)

    if row.endswith("~"):
        # We determine the most specific regex for the row at matching in match_row_to_acls
        row = row[:-1] + "(.+)"
    elif row.endswith("..."):
        row = row[:-3]
    elif "~/" in row:
        # ~/{regex}/ -> {regex}, () не нужны поскольку уже (?:) - non-captured
        row = re.sub(r"~/(((?!~/).)+)/", r"\1", row)
    else:
        row += r"(?:\s|$)"
    row = re.sub(r"\s+", r"\\s+", row)
    return re.compile("^" + row, flags=flags)


# =====
def _split_rows(text):
    for row in re.split(r"\n(?!\s*%(?!context))", text):
        yield row.replace("\n", " ")


def _parse_tree_with_params(raw_tree, scheme, context=None):
    tree = odict()
    if context is None:
        context = {}
    for (raw_rule, children) in raw_tree.items():
        (row, params) = _parse_raw_rule(raw_rule, scheme)
        row_type = "normal"

        if row.startswith("!"):
            row = row[1:].strip()
            if len(row) == 0:
                continue
            row_type = "ignore"
        elif context_raw := params.get("context"):
            context = _parse_context(context, context_raw)
            continue
        tree[raw_rule] = {
            "row": row,
            "type": row_type,
            "params": params,
            "children": _parse_tree_with_params(children, scheme, context.copy()),
            "raw_rule": raw_rule,
            "context": context.copy(),
        }
    return tree


def _parse_raw_rule(raw_rule, scheme):
    try:
        index = raw_rule.index("%")
        params = {
            key: (value.strip() if len(value.strip()) != 0 else "1")
            for (key, value) in re.findall(r"%([a-zA-Z_]\w*)(?:=([^%]*))?", raw_rule[index:])
        }
        if params:
            raw_rule = raw_rule[:index].strip()
    except ValueError:
        params = {}

    row = re.sub(r"\s+", " ", raw_rule.strip())
    params = _fill_and_validate(params, scheme, raw_rule)
    return row, params


def _fill_and_validate(params, scheme, raw_rule):
    return {
        key: (attrs["validator"](params[key]) if key in params else (
            attrs["default"](raw_rule) if callable(attrs["default"]) else attrs["default"]
        ))
        for (key, attrs) in scheme.items()
    }


def match_context(ifcontext, context):
    if not ifcontext:
        return True
    for ifcontext_value in ifcontext:
        name, value = ifcontext_value.split(":")
        if name in context:
            if context[name] == value:
                return True
    return False


def _parse_context(context, row):
    name, value = row.strip().split(":")
    return lib.merge_dicts(context, {name: value})



import abc
import inspect
import os
import re
import sys
import time
import warnings
from collections import OrderedDict as odict
from itertools import chain, groupby
from operator import itemgetter
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Set,
    Tuple,
    Union,
    cast,
)

import colorama
from contextlog import get_logger

import annet.deploy
import annet.deploy_ui
import annet.lib
from annet import cli_args
from annet import diff as ann_diff
from annet import filtering
from annet import gen as ann_gen
from annet import patching, rulebook, tracing
from annet.vendors import tabparser
from annet.annlib import jsontools
from annet.annlib.netdev.views.hardware import HardwareView
from annet.annlib.types import GeneratorType
from annet.deploy import DeployDriver, Fetcher
from annet.diff import file_differ_connector
from annet.filtering import Filterer
from annet.hardware import hardware_connector
from annet.output import format_file_diff, output_driver_connector, print_err_label
from annet.parallel import Parallel, TaskResult
from annet.reference import RefTracker
from annet.rulebook import deploying
from annet.storage import Device, get_storage
from annet.types import Diff, ExitCode, OldNewResult, Op, PCDiff
from annet.vendors import registry_connector


DEFAULT_INDENT = "  "


def patch_from_pre(pre, hw, rb, add_comments, ref_track=None, do_commit=True):
    if not ref_track:
        ref_track = RefTracker()
    orderer = patching.Orderer(rb["ordering"], hw.vendor)
    orderer.ref_insert(ref_track)
    return patching.make_patch(
        pre=pre,
        rb=rb,
        hw=hw,
        add_comments=add_comments,
        orderer=orderer,
        do_commit=do_commit,
    )


def _diff_and_patch(
    device, old, new, acl_rules, filter_acl_rules,
    add_comments, ref_track=None, do_commit=True, rb=None
) -> Tuple[Diff, Dict]:
    if rb is None:
        rb = rulebook.get_rulebook(device.hw)
    # [NOCDEV-5532] Передаем в diff только релевантные для logic'a части конфига
    if acl_rules is not None:
        old = patching.apply_acl(old, acl_rules)
        new = patching.apply_acl(new, acl_rules, with_annotations=add_comments)

    diff_tree = patching.make_diff(old, new, rb, [acl_rules, filter_acl_rules])
    pre = patching.make_pre(diff_tree)
    patch_tree = patch_from_pre(pre, device.hw, rb, add_comments, ref_track, do_commit)
    diff_tree = patching.strip_unchanged(diff_tree)

    return (diff_tree, patch_tree)


# =====
def _read_old_new_diff_patch(old: Dict[str, Dict], new: Dict[str, Dict], hw: HardwareView, add_comments: bool):
    rb = rulebook.get_rulebook(hw)
    diff_obj = patching.make_diff(old, new, rb, [])
    diff_obj = patching.strip_unchanged(diff_obj)
    pre = patching.make_pre(diff_obj)
    patchtree = patch_from_pre(pre, hw, rb, add_comments)
    return rb, diff_obj, pre, patchtree


def _read_old_new_configs(old_path, new_path):
    _logger = get_logger()
    with open(old_path) as f:
        _logger.debug("Reading %r ...", old_path)
        old_config = f.read()

    with open(new_path) as f:
        _logger.debug("Reading %r ...", new_path)
        new_config = f.read()

    return old_config, new_config


def _read_old_new_hw(old_path: str, old_config: str, new_path: str, new_config: str, args: cli_args.FileInputOptions):
    _logger = get_logger()

    hw = args.hw
    if isinstance(args.hw, str):
        hw = HardwareView(args.hw, "")

    try:
        old, old_hw, old_score = _parse_device_config(old_config, hw)
    except tabparser.ParserError:
        _logger.exception("Parser error: %r", old_path)
        raise

    try:
        new, new_hw, new_score = _parse_device_config(new_config, hw)
    except tabparser.ParserError:
        _logger.exception("Parser error: %r", new_path)
        raise

    hw = new_hw
    if old_score > new_score:
        hw = old_hw
    if old_hw != new_hw:
        _logger.warning("Old and new detected hw differs, assume %r", hw)

    return old, new, hw


@tracing.function
def _read_old_new_cfgdumps(args: cli_args.FileInputOptions):
    _logger = get_logger()
    old_path, new_path = os.path.normpath(args.old), os.path.normpath(args.new)
    if not os.path.isdir(old_path):
        yield (old_path, new_path)
        return
    _logger.info("Scanning cfgdumps: %s/*.cfg ...", old_path)
    cfgdump_reg = re.compile(r"^[^\s]+\.cfg$")
    if os.path.isdir(old_path) and os.path.isdir(new_path):
        if cfgdump_reg.match(os.path.basename(old_path)) and cfgdump_reg.match(os.path.basename(new_path)):
            yield (old_path, new_path)
    for name in os.listdir(old_path):
        old_path_name = os.path.join(old_path, name)
        new_path_name = os.path.join(new_path, name)
        if not os.path.exists(new_path_name):
            _logger.debug("Ignoring file %s: not exist %s", name, new_path_name)
            continue
        yield (old_path_name, new_path_name)


def _parse_device_config(text, hw):
    score = 1
    vendor_registry = registry_connector.get()

    if not hw:
        hw, score = guess_hw(text)

    config = tabparser.parse_to_tree(
        text=text,
        splitter=vendor_registry.match(hw).make_formatter().split,
    )

    return config, hw, score


# =====
def _format_patch_blocks(patch_tree, hw, indent):
    formatter = registry_connector.get().match(hw).make_formatter(indent=indent)
    return formatter.patch(patch_tree)


# =====
def _print_pre_as_diff(pre, show_rules, indent, file=None, _level=0):
    for (raw_rule, content) in sorted(pre.items(), key=itemgetter(0)):
        rule_printed = False
        for (op, sign) in [  # FIXME: Not very effective
            (Op.REMOVED, colorama.Fore.RED + "-"),
            (Op.ADDED, colorama.Fore.GREEN + "+"),
            (Op.AFFECTED, colorama.Fore.CYAN + " "),
        ]:
            items = content["items"].items()
            for (_, diff) in items:  # pylint: disable=redefined-outer-name
                if show_rules and not rule_printed and not raw_rule == "__MULTILINE_BODY__":
                    print("%s%s# %s%s%s" % (colorama.Style.BRIGHT, colorama.Fore.BLACK, (indent * _level),
                                            raw_rule, colorama.Style.RESET_ALL), file=file)
                    rule_printed = True
                for item in sorted(diff[op], key=itemgetter("row")):
                    print("%s%s%s %s%s" % (colorama.Style.BRIGHT, sign, (indent * _level),
                                           item["row"], colorama.Style.RESET_ALL), file=file)
                    if len(item["children"]) != 0:
                        _print_pre_as_diff(item["children"], show_rules, indent, file, _level + 1)
                        rule_printed = False


class PoolProgressLogger:
    def __init__(self, device_fqdns: Dict[int, str]):
        self.device_fqdns = device_fqdns

    def __call__(self, pool: Parallel, task_result: TaskResult):
        progress_logger = get_logger("progress")
        perc = int(pool.tasks_done / len(self.device_fqdns) * 100)

        fqdn = self.device_fqdns[task_result.device_id]
        elapsed_time = "%dsec" % int(time.monotonic() - task_result.extra["start_time"])
        if task_result.extra.get("regression", False):
            status = task_result.extra["status"]
            status_color = task_result.extra["status_color"]
            message = task_result.extra["message"]
        else:
            status = "OK" if task_result.exc is None else "FAIL"
            status_color = colorama.Fore.GREEN if status == "OK" else colorama.Fore.RED
            message = "" if status == "OK" else str(task_result.exc)
        progress_logger.info(message,
                             perc=perc, fqdn=fqdn, status=status, status_color=status_color,
                             worker=task_result.worker_name, task_time=elapsed_time)
        return task_result


def log_host_progress_cb(pool: Parallel, task_result: TaskResult):
    warnings.warn(
        "log_host_progress_cb is deprecated, use PoolProgressLogger",
        DeprecationWarning,
        stacklevel=2,
    )
    args = cast(cli_args.QueryOptions, pool.args[0])
    connector = get_storage()
    storage_opts = connector.opts().parse_params({}, args)
    with connector.storage()(storage_opts) as storage:
        fqdns = storage.resolve_fdnds_by_query(args.query)
    PoolProgressLogger(device_fqdns=fqdns)(pool, task_result)


# =====
def gen(args: cli_args.ShowGenOptions, loader: ann_gen.Loader):
    """ Сгенерировать конфиг для устройств """
    stdin = args.stdin(filter_acl=args.filter_acl, config=None)

    filterer = filtering.filterer_connector.get()
    pool = Parallel(ann_gen.worker, args, stdin, loader, filterer).tune_args(args)
    if args.show_hosts_progress:
        pool.add_callback(PoolProgressLogger(loader.device_fqdns))

    return pool.run(loader.device_ids, args.tolerate_fails, args.strict_exit_code)


# =====


def patch(args: cli_args.ShowPatchOptions, loader: ann_gen.Loader):
    """ Сгенерировать патч для устройств """
    if args.config == "running":
        fetcher = annet.deploy.get_fetcher()
        ann_gen.live_configs = annet.lib.do_async(fetcher.fetch(loader.devices, processes=args.parallel))
    stdin = args.stdin(filter_acl=args.filter_acl, config=args.config)

    filterer = filtering.filterer_connector.get()
    pool = Parallel(_patch_worker, args, stdin, loader, filterer).tune_args(args)
    if args.show_hosts_progress:
        pool.add_callback(PoolProgressLogger(loader.device_fqdns))
    return pool.run(loader.device_ids, args.tolerate_fails, args.strict_exit_code)


def _patch_worker(device_id, args: cli_args.ShowPatchOptions, stdin, loader: ann_gen.Loader, filterer: filtering.Filterer):
    for res, _, patch_tree in res_diff_patch(device_id, args, stdin, loader, filterer):
        old_files = res.old_files
        new_files = res.get_new_files(args.acl_safe)
        new_json_fragment_files = res.get_new_file_fragments(args.acl_safe)
        if new_files:
            for path, (cfg_text, _cmds) in new_files.items():
                label = res.device.hostname + os.sep + path
                if old_files.get(path) != cfg_text:
                    yield label, cfg_text, False
        elif res.old_json_fragment_files or new_json_fragment_files:
            for path, (new_json_cfg, _cmds) in new_json_fragment_files.items():
                label = res.device.hostname + os.sep + path
                old_json_cfg = res.old_json_fragment_files[path]
                json_patch = jsontools.make_patch(old_json_cfg, new_json_cfg)
                yield (
                    label,
                    jsontools.format_json(json_patch),
                    False,
                )
        elif patch_tree:
            yield (
                "%s.patch" % res.device.hostname,
                _format_patch_blocks(patch_tree, res.device.hw, args.indent),
                False,
            )


# =====
def res_diff_patch(
        device_id, args: cli_args.ShowPatchOptions, stdin, loader: ann_gen.Loader, filterer: filtering.Filterer,
) -> Iterable[Tuple[OldNewResult, Dict, Dict]]:
    for res in ann_gen.old_new(
        args,
        config=args.config,
        loader=loader,
        filterer=filterer,
        stdin=stdin,
        device_ids=[device_id],
        no_new=args.clear,
        do_files_download=True,
    ):
        old = res.get_old(args.acl_safe)
        new = res.get_new(args.acl_safe)
        new_json_fragment_files = res.get_new_file_fragments(args.acl_safe)

        device = res.device
        acl_rules = res.get_acl_rules(args.acl_safe)
        if res.old_json_fragment_files or new_json_fragment_files:
            yield res, None, None
        elif old is not None:
            (diff_tree, patch_tree) = _diff_and_patch(device, old, new, acl_rules, res.filter_acl_rules,
                                                      args.add_comments)
            yield res, diff_tree, patch_tree


def diff(
    args: cli_args.DiffOptions,
    loader: ann_gen.Loader,
    device_ids: List[Any]
) -> tuple[Mapping[Device, Union[Diff, PCDiff]], Mapping[Device, Exception]]:
    """ Сгенерировать дифф для устройств """
    if args.config == "running":
        fetcher = annet.deploy.get_fetcher()
        ann_gen.live_configs = annet.lib.do_async(
            fetcher.fetch(
                [device for device in loader.devices if device.id in device_ids],
                processes=args.parallel
            ),
            new_thread=True
        )
    stdin = args.stdin(filter_acl=args.filter_acl, config=None)

    filterer = filtering.filterer_connector.get()
    pool = Parallel(ann_diff.worker, args, stdin, loader, filterer).tune_args(args)
    if args.show_hosts_progress:
        fqdns = {k: v for k, v in loader.device_fqdns.items() if k in device_ids}
        pool.add_callback(PoolProgressLogger(fqdns))

    return pool.run(device_ids, args.tolerate_fails, args.strict_exit_code)


def collapse_texts(texts: Mapping[str, str | Generator[str, None, None]]) -> Mapping[Tuple[str, ...], str]:
    """
    Группировка текстов.
    :param texts:
    :return: словарь с несколькими хостнеймами в ключе.
    """
    diffs_with_orig = {}
    for key, value in texts.items():
        if inspect.isgenerator(value):
            lines = list(value)
            diffs_with_orig[key] = ["".join(lines), lines]
        else:
            diffs_with_orig[key] = [value, value.splitlines()]

    res = {}
    for _, collapsed_diff_iter in groupby(
        sorted(diffs_with_orig.items(), key=lambda x: (x[0], x[1][1])),
        key=lambda x: x[1][1]
    ):
        collapsed_diff = list(collapsed_diff_iter)
        res[tuple(x[0] for x in collapsed_diff)] = collapsed_diff[0][1][0]

    return res


class DeployerJob(abc.ABC):
    def __init__(self, device, args: cli_args.DeployOptions):
        self.args = args
        self.device = device
        self.add_comments = False
        self.diff_lines = []
        self.cmd_lines: List[str] = []
        self.deploy_cmds = odict()
        self.diffs = {}
        self.failed_configs = {}
        self._has_diff = False

    @abc.abstractmethod
    def parse_result(self, res):
        pass

    def collapseable_diffs(self):
        return {}

    def has_diff(self):
        return self._has_diff

    @staticmethod
    def from_device(device, args: cli_args.DeployOptions):
        if device.hw.vendor == "pc":
            return PCDeployerJob(device, args)
        return CliDeployerJob(device, args)


class CliDeployerJob(DeployerJob):
    def parse_result(self, res: OldNewResult):
        device = res.device
        old = res.get_old(self.args.acl_safe)
        new = res.get_new(self.args.acl_safe)
        acl_rules = res.get_acl_rules(self.args.acl_safe)
        err = res.err

        if err:
            self.failed_configs[device.fqdn] = err
            return

        (diff_obj, patch_tree) = _diff_and_patch(device, old, new, acl_rules,
                                                 res.filter_acl_rules, self.add_comments,
                                                 do_commit=not self.args.dont_commit)

        formatter = registry_connector.get().match(device.hw).make_formatter(indent="")
        cmds = formatter.cmd_paths(patch_tree)
        if not cmds:
            return

        self._has_diff = True
        self.diffs[device] = diff_obj
        self.cmd_lines.extend(["= %s " % device.hostname, ""])
        self.cmd_lines.extend(map(itemgetter(-1), cmds))
        self.cmd_lines.append("")
        deployer_driver = annet.deploy.get_deployer()
        self.deploy_cmds[device] = deployer_driver.apply_deploy_rulebook(
            device.hw, cmds,
            do_commit=not self.args.dont_commit
        )
        for cmd in deployer_driver.build_exit_cmdlist(device.hw):
            self.deploy_cmds[device].add_cmd(cmd)

    def collapseable_diffs(self):
        return self.diffs


class PCDeployerJob(DeployerJob):
    def parse_result(self, res: ann_gen.OldNewResult):
        device = res.device
        old_files = res.old_files
        new_files = res.get_new_files(self.args.acl_safe)
        old_json_fragment_files = res.old_json_fragment_files
        new_json_fragment_files = res.get_new_file_fragments(self.args.acl_safe)
        err = res.err

        if err:
            self.failed_configs[device.fqdn] = err
            return
        elif not new_files and not new_json_fragment_files:
            return

        enable_reload = self.args.entire_reload is not cli_args.EntireReloadFlag.no
        force_reload = self.args.entire_reload is cli_args.EntireReloadFlag.force

        upload_files: Dict[str, bytes] = {}
        reload_cmds: Dict[str, bytes] = {}
        generator_types: Dict[str, GeneratorType] = {}
        differ = file_differ_connector.get()
        for generator_type, pc_files in [(GeneratorType.ENTIRE, new_files), (GeneratorType.JSON_FRAGMENT, new_json_fragment_files)]:
            for file, (file_content_or_json_cfg, cmds) in pc_files.items():
                if generator_type == GeneratorType.ENTIRE:
                    file_content: str = file_content_or_json_cfg
                    diff_content = "\n".join(differ.diff_file(res.device.hw, file, old_files.get(file), file_content))
                else:  # generator_type == GeneratorType.JSON_FRAGMENT
                    old_json_cfg = old_json_fragment_files[file]
                    json_patch = jsontools.make_patch(old_json_cfg, file_content_or_json_cfg)
                    file_content = jsontools.format_json(json_patch)
                    old_text = jsontools.format_json(old_json_cfg)
                    new_text = jsontools.format_json(file_content_or_json_cfg)
                    diff_content = "\n".join(differ.diff_file(res.device.hw, file, old_text, new_text))

                if diff_content or force_reload:
                    self._has_diff |= True

                    upload_files[file] = file_content.encode()
                    generator_types[file] = generator_type
                    self.cmd_lines.append("= %s/%s " % (device.hostname, file))
                    self.cmd_lines.extend([file_content, ""])
                    self.diff_lines.append("= %s/%s " % (device.hostname, file))
                    self.diff_lines.extend([diff_content, ""])

                    if enable_reload:
                        reload_cmds[file] = cmds.encode()
                        self.cmd_lines.append("= Deploy cmds %s/%s " % (device.hostname, file))
                        self.cmd_lines.extend([cmds, ""])

        if self._has_diff:
            self.deploy_cmds[device] = {
                "files": upload_files,
                "cmds": reload_cmds,
                "generator_types": generator_types,
            }
            self.diffs[device] = upload_files
            deployer_driver = annet.deploy.get_deployer()
            before, after = deployer_driver.build_configuration_cmdlist(device.hw)
            for cmd in deployer_driver.build_exit_cmdlist(device.hw):
                after.add_cmd(cmd)
            cmds_pre_files = {}
            rules = rulebook.get_rulebook(device.hw)["deploying"]
            for file, content in self.deploy_cmds[device]["files"].items():
                rule = deploying.match_deploy_rule(rules, [file], content)
                before_more, after_more = annet.deploy.make_apply_commands(rule, res.device.hw, do_commit=True, do_finalize=True, path=file)

                cmds_pre_files[file] = "\n".join(map(str, chain(before, before_more))).encode(encoding="utf-8")
                after_cmds = "\n".join(map(str, chain(after, after_more))).encode(encoding="utf-8")
                if after_cmds:
                    self.deploy_cmds[device]["cmds"][file] += b"\n" + after_cmds
            self.deploy_cmds[device]["cmds_pre_files"] = cmds_pre_files


class Deployer:
    def __init__(self, args: cli_args.DeployOptions):
        self.args = args

        self.cmd_lines = []
        self.deploy_cmds = odict()
        self.diffs = {}
        self.failed_configs: Dict[str, Exception] = {}
        self.fqdn_to_device: Dict[str, Device] = {}
        self.empty_diff_hostnames: Set[str] = set()

        self._collapseable_diffs = {}
        self._diff_lines: List[str] = []

    def parse_result(self, job: DeployerJob, result: ann_gen.OldNewResult):
        logger = get_logger(job.device.hostname)

        job.parse_result(result)
        self.failed_configs.update(job.failed_configs)

        if job.has_diff():
            self.cmd_lines.extend(job.cmd_lines)
            self.deploy_cmds.update(job.deploy_cmds)
            self.diffs.update(job.diffs)

            self.fqdn_to_device[result.device.fqdn] = result.device
            self._collapseable_diffs.update(job.collapseable_diffs())
            self._diff_lines.extend(job.diff_lines)
        else:
            logger.info("empty diff")

    def diff_lines(self) -> List[str]:
        diff_lines = []
        diff_lines.extend(self._diff_lines)
        for devices, diff_obj in ann_diff.collapse_diffs(self._collapseable_diffs).items():
            if not diff_obj:
                self.empty_diff_hostnames.update(dev.hostname for dev in devices)
            if not self.args.no_ask_deploy:
                # разобьём список устройств на несколько линий
                dest_name = ""
                try:
                    _, term_columns_str = os.popen("stty size", "r").read().split()
                    term_columns = int(term_columns_str)
                except Exception:
                    term_columns = 2 ** 32
                fqdns = [dev.hostname for dev in devices]
                while fqdns:
                    fqdn = fqdns.pop()
                    if len(dest_name) == 0:
                        dest_name = "= %s" % fqdn
                    elif len(dest_name) + len(fqdn) < term_columns:
                        dest_name = "%s, %s" % (dest_name, fqdn)
                    else:
                        diff_lines.extend([dest_name])
                        dest_name = "= %s" % fqdn
                    if not fqdns:
                        diff_lines.extend([dest_name, ""])
            else:
                dest_name = "= %s" % ", ".join([dev.hostname for dev in devices])
                diff_lines.extend([dest_name, ""])

            for line in registry_connector.get().match(devices[0].hw).make_formatter().diff(diff_obj):
                diff_lines.append(line)
            diff_lines.append("")
        return diff_lines

    def ask_deploy(self) -> str:
        return self._ask("y", annet.deploy_ui.AskConfirm(
            text="\n".join(self.diff_lines()),
            alternative_text="\n".join(self.cmd_lines),
        ))

    def ask_rollback(self) -> str:
        return self._ask("n", annet.deploy_ui.AskConfirm(
            text="Execute rollback?\n",
            alternative_text="",
        ))

    def _ask(self, default_ans: str, ask: annet.deploy_ui.AskConfirm) -> str:
        # если filter_acl из stdin то с ним уже не получится работать как с терминалом
        ans = default_ans
        if not self.args.no_ask_deploy:
            try:
                if not os.isatty(sys.stdin.fileno()):
                    pts_path = os.ttyname(sys.stdout.fileno())
                    pts = open(pts_path, "r")  # pylint: disable=consider-using-with
                    os.dup2(pts.fileno(), sys.stdin.fileno())
            except OSError:
                pass
            ans = ask.loop()
        return ans

    def check_diff(self, result: annet.deploy.DeployResult, loader: ann_gen.Loader):
        diff_args = self.args.copy_from(
            self.args,
            config="running",
        )
        if not diff_args.query:
            return

        # clear cache
        ann_gen.live_configs = None

        # collect new diffs for devices on which we had successfully uploaded something
        success_device_ids = []
        for host, hres in result.results.items():
            if (
                not isinstance(hres, Exception) and
                host not in self.empty_diff_hostnames
            ):
                device = self.fqdn_to_device[host]
                success_device_ids.append(device.id)
        diffs, failed = diff(diff_args, loader, success_device_ids)
        for device_id, exc in failed.items():
            self.failed_configs[loader.get_device(device_id).fqdn] = exc

        # "collapse" non-PC diffs
        diffs_by_device_id = ann_diff.collapse_diffs({
            loader.get_device(device_id): diff
            for device_id, diff in diffs.items()
            if diff and not isinstance(diff, PCDiff)
        })
        # add PC diffs as is
        diffs_by_device_id.update({
            (loader.get_device(device_id),): diff
            for device_id, diff in diffs.items()
            if diff and isinstance(diff, PCDiff)
        })
        if diffs_by_device_id:
            print("The diff is still present:")
            for devices, diff_obj in diffs_by_device_id.items():
                for dev in devices:
                    self.failed_configs[dev.fqdn] = Warning("Deploy OK, but diff still exists")
                if isinstance(diff_obj, PCDiff):
                    for diff_file in diff_obj.diff_files:
                        print_err_label(diff_file.label)
                        print("\n".join(format_file_diff(diff_file.diff_lines)))
                else:
                    output_driver = output_driver_connector.get()
                    dest_name = ", ".join([output_driver.cfg_file_names(dev)[0] for dev in devices])
                    print_err_label(dest_name)
                    _print_pre_as_diff(patching.make_pre(diff_obj), diff_args.show_rules, diff_args.indent)


def deploy(
    args: cli_args.DeployOptions,
    loader: ann_gen.Loader,
    deployer: Deployer,
    filterer: Filterer,
    fetcher: Fetcher,
    deploy_driver: DeployDriver,
) -> ExitCode:
    return annet.lib.do_async(adeploy(args, loader, deployer, filterer, fetcher, deploy_driver))


async def adeploy(
    args: cli_args.DeployOptions,
    loader: ann_gen.Loader,
    deployer: Deployer,
    filterer: Filterer,
    fetcher: Fetcher,
    deploy_driver: DeployDriver,
) -> ExitCode:
    """ Сгенерировать конфиг для устройств и задеплоить его """
    ret: ExitCode = 0
    ann_gen.live_configs = await fetcher.fetch(devices=loader.devices, processes=args.parallel, max_slots=args.max_slots)

    device_ids = [d.id for d in loader.devices]
    for res in ann_gen.old_new(
        args,
        config=args.config,
        loader=loader,
        no_new=args.clear,
        stdin=args.stdin(filter_acl=args.filter_acl, config=args.config),
        do_files_download=True,
        device_ids=device_ids,
        filterer=filterer,
    ):
        # Меняем exit code если хоть один device ловил exception
        if res.err is not None:
            if not args.tolerate_fails:
                raise res.err
            get_logger(res.device.hostname).error("error generating configs", exc_info=res.err)
            ret |= 2 ** 3
        job = DeployerJob.from_device(res.device, args)
        deployer.parse_result(job, res)

    deploy_cmds = deployer.deploy_cmds
    result = annet.deploy.DeployResult(hostnames=[], results={}, durations={}, original_states={})
    if deploy_cmds:
        ans = deployer.ask_deploy()
        if ans != "y":
            return 2 ** 2

        if sys.stdout.isatty() and not args.no_progress:
            progress_bar = annet.deploy_ui.ProgressBars(odict([(device.fqdn, {}) for device in deploy_cmds]))
            progress_bar.init()
            with progress_bar:
                progress_bar.start_terminal_refresher()
                result = await deploy_driver.bulk_deploy(deploy_cmds, args, progress_bar=progress_bar)
                await progress_bar.wait_for_exit()
            progress_bar.screen.clear()
            progress_bar.stop_terminal_refresher()
        else:
            result = await deploy_driver.bulk_deploy(deploy_cmds, args)

    rolled_back = False
    rollback_cmds = {deployer.fqdn_to_device[x]: cc for x, cc in result.original_states.items() if cc}
    if args.rollback and rollback_cmds:
        ans = deployer.ask_rollback()
        if rollback_cmds and ans == "y":
            rolled_back = True
            await deploy_driver.bulk_deploy(rollback_cmds, args)

    if not args.no_check_diff and not rolled_back:
        deployer.check_diff(result, loader)

    if deployer.failed_configs:
        result.add_results(deployer.failed_configs)
        ret |= 2 ** 1

    annet.deploy.show_bulk_report(result.hostnames, result.results, result.durations, log_dir=None)
    for host_result in result.results.values():
        if isinstance(host_result, Exception):
            ret |= 2 ** 0
            break
    return ret


def file_diff(args: cli_args.FileDiffOptions):
    """ Создать дифф по рулбуку между файлами или каталогами """
    old_new = list(_read_old_new_cfgdumps(args))
    pool = Parallel(file_diff_worker, args).tune_args(args)
    return pool.run(old_new, tolerate_fails=True)


def file_diff_worker(
    old_new: Tuple[str, str],
    args: cli_args.FileDiffOptions
) -> Generator[Tuple[str, str, bool], None, None]:
    old_path, new_path = old_new
    hw = args.hw
    if isinstance(args.hw, str):
        hw = HardwareView(args.hw, "")

    if os.path.isdir(old_path) and os.path.isdir(new_path):
        hostname = os.path.basename(new_path)
        new_files = {
            relative_cfg_path: (cfg_text, "")
            for relative_cfg_path, cfg_text in ann_gen.load_pc_config(new_path).items()
        }
        old_files = ann_gen.load_pc_config(old_path)
        for diff_file in ann_diff.pc_diff(hw, hostname, old_files, new_files):
            diff_text = (
                "\n".join(diff_file.diff_lines)
                if args.no_color
                else "\n".join(format_file_diff(diff_file.diff_lines))
            )
            if diff_text:
                yield diff_file.label, diff_text, False
    else:
        old_config, new_config = _read_old_new_configs(old_path, new_path)
        if old_config == new_config:
            return

        old, new, hw = _read_old_new_hw(old_path, old_config, new_path, new_config, args)
        _, __, pre, ___ = _read_old_new_diff_patch(old, new, hw, add_comments=False)

        if diff_lines := ann_diff.gen_pre_as_diff(pre, args.show_rules, args.indent, args.no_color):
            yield os.path.basename(new_path), "".join(diff_lines), False


@tracing.function
def file_patch(args: cli_args.FilePatchOptions):
    """ Создать патч между файлами или каталогами """
    old_new = list(_read_old_new_cfgdumps(args))
    pool = Parallel(file_patch_worker, args).tune_args(args)
    return pool.run(old_new, tolerate_fails=True)


def file_patch_worker(old_new: Tuple[str, str], args: cli_args.FileDiffOptions) -> Generator[
    Tuple[str, str, bool], None, None
]:
    old_path, new_path = old_new
    if os.path.isdir(old_path) and os.path.isdir(new_path):
        for relative_cfg_path, cfg_text in ann_gen.load_pc_config(new_path).items():
            label = os.path.join(os.path.basename(new_path), relative_cfg_path)
            yield label, cfg_text, False
    else:
        old_config, new_config = _read_old_new_configs(old_path, new_path)
        if old_config == new_config:
            return

        old, new, hw = _read_old_new_hw(old_path, old_config, new_path, new_config, args)
        _, __, ___, patch_tree = _read_old_new_diff_patch(old, new, hw, args.add_comments)
        patch_text = _format_patch_blocks(patch_tree, hw, args.indent)
        if patch_text:
            yield os.path.basename(new_path), patch_text, False


def guess_hw(config_text: str):
    """Пытаемся угадать вендора и hw на основе
    текста конфига и annet/rulebook/texts/*.rul"""
    scores = {}
    hw_provider = hardware_connector.get()
    vendor_registry = registry_connector.get()
    for vendor in vendor_registry:
        hw = hw_provider.vendor_to_hw(vendor)
        rb = rulebook.get_rulebook(hw)
        fmtr = vendor_registry[vendor].make_formatter()

        try:
            config = tabparser.parse_to_tree(config_text, fmtr.split)
        except Exception:
            continue

        pre = patching.make_pre(patching.make_diff({}, config, rb, []))
        metric = _count_pre_score(pre)
        scores[metric] = hw

    if not scores:
        raise RuntimeError("No formatter was guessed")

    max_score = max(scores.keys())
    hw = scores[max_score]

    return hw, max_score


def _count_pre_score(top_pre) -> float:
    """Обходим вширь pre-конфиг
    и подсчитываем количество заматчившихся
    правил на каждом из уровней.

    Чем больше результирующий приоритет
    тем больше рулбук соответсвует конфигу.
    """
    score = 0
    scores = []
    cur, child = [top_pre], []
    while cur:
        for pre in cur.pop().values():
            score += 1
            for item in pre["items"].values():
                for op in [Op.ADDED, Op.AFFECTED, Op.REMOVED]:
                    child += [x["children"] for x in item[op]]
        if not cur:
            scores.append(score)
            score = 0
            cur, child = child, []
    result = 0
    for i in reversed(scores):
        result <<= i.bit_length()
        result += i
    if result > 0:
        result = 1 - (1 / result)
    return float(result)



from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence, Iterable
from typing import Any, cast, Literal

from annet.generators import PartialGenerator
from annet.rpl import (
    CommunityActionValue,
    ResultType, RoutingPolicyStatement, RoutingPolicy, ConditionOperator, SingleCondition, SingleAction, ActionType,
    MatchField, PrefixMatchValue, AndCondition, Action,
)
from annet.rpl.statement_builder import AsPathActionValue, NextHopActionValue, ThenField
from annet.rpl_generators.entities import (
    arista_well_known_community,
    CommunityList, RDFilter, PrefixListNameGenerator, CommunityLogic, mangle_united_community_list_name,
    IpPrefixList, group_community_members, CommunityType, JuniperPrefixListNameGenerator
)

HUAWEI_MATCH_COMMAND_MAP: dict[str, str] = {
    MatchField.as_path_filter: "as-path-filter {option_value}",
    MatchField.metric: "cost {option_value}",
    MatchField.protocol: "protocol {option_value}",
    MatchField.interface: "interface {option_value}",
}

HUAWEI_THEN_COMMAND_MAP: dict[str, str] = {
    ThenField.local_pref: "local-preference {option_value}",
    ThenField.metric_type: "cost-type {option_value}",
    ThenField.mpls_label: "mpls-label",
    ThenField.origin: "origin {option_value}",
    ThenField.tag: "tag {option_value}",
    # unsupported: resolution
    # unsupported: rpki_valid_state
}
HUAWEI_RESULT_MAP = {
    ResultType.ALLOW: "permit",
    ResultType.DENY: "deny",
    ResultType.NEXT: "permit"
}
ARISTA_RESULT_MAP = {
    ResultType.ALLOW: "permit",
    ResultType.DENY: "deny",
    ResultType.NEXT: "permit"
}
ARISTA_MATCH_COMMAND_MAP: dict[str, str] = {
    MatchField.interface: "interface {option_value}",
    MatchField.metric: "metric {option_value}",
    MatchField.as_path_filter: "as-path {option_value}",
    MatchField.protocol: "source-protocol {option_value}",
    # unsupported: rd
}
ARISTA_THEN_COMMAND_MAP: dict[str, str] = {
    ThenField.local_pref: "local-preference {option_value}",
    ThenField.origin: "origin {option_value}",
    ThenField.tag: "tag {option_value}",
    ThenField.metric_type: "metric-type {option_value}",
    # unsupported: mpls_label
    # unsupported: resolution
    # unsupported: rpki_valid_state
}
IOSXR_MATCH_COMMAND_MAP: dict[str, str] = {
    MatchField.as_path_filter: "as-path in {option_value}",
    MatchField.protocol: "protocol is {option_value}",
    # unsupported: interface
    # unsupported: metric
    # unsupported: large-community
}
IOSXR_THEN_COMMAND_MAP: dict[str, str] = {
    ThenField.local_pref: "local-preference {option_value}",
    ThenField.origin: "origin {option_value}",
    ThenField.tag: "tag {option_value}",
    ThenField.metric_type: "metric-type {option_value}",
    # unsupported: mpls_label  # label?
    # unsupported: resolution
    # unsupported: rpki_valid_state
    # unsupported: large_community
}
IOSXR_RESULT_MAP = {
    ResultType.ALLOW: "done",
    ResultType.DENY: "drop",
    ResultType.NEXT: "pass"
}
JUNIPER_MATCH_COMMAND_MAP: dict[str, str] = {
    MatchField.protocol: "protocol {option_value}",
    MatchField.metric: "metric {option_value}",
    MatchField.as_path_filter: "as-path {option_value}",
    MatchField.local_pref: "local-preference {option_value}",
    # unsupported: rd
    # unsupported: interface
    # unsupported: net_len
    # unsupported: family
}
JUNIPER_THEN_COMMAND_MAP: dict[str, str] = {
    ThenField.local_pref: "local-preference {option_value}",
    ThenField.origin: "origin {option_value}",
    ThenField.tag: "tag {option_value}",
    ThenField.metric: "metric {option_value}",
    # unsupported: rpki_valid_state
    # unsupported: resolution
    # unsupported: mpls_label
    # unsupported: metric_type
}
JUNIPER_RESULT_MAP = {
    ResultType.ALLOW: "accept",
    ResultType.DENY: "reject",
    ResultType.NEXT: "next term"
}


class RoutingPolicyGenerator(PartialGenerator, ABC):
    TAGS = ["policy", "rpl", "routing"]

    @abstractmethod
    def get_prefix_lists(self, device: Any) -> list[IpPrefixList]:
        raise NotImplementedError()

    @abstractmethod
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        raise NotImplementedError()

    @abstractmethod
    def get_community_lists(self, device: Any) -> list[CommunityList]:
        raise NotImplementedError()

    @abstractmethod
    def get_rd_filters(self, device: Any) -> list[RDFilter]:
        raise NotImplementedError()

    # huawei
    def acl_huawei(self, _):
        return r"""
        route-policy *
            ~ %global=1
        """

    def _huawei_match(
            self,
            device: Any,
            condition: SingleCondition[Any],
            communities: dict[str, CommunityList],
            rd_filters: dict[str, RDFilter],
            name_generator: PrefixListNameGenerator,
    ) -> Iterator[Sequence[str]]:
        if condition.field == MatchField.community:
            if condition.operator is ConditionOperator.HAS:
                if len(condition.value) > 1:
                    raise NotImplementedError("Multiple HAS for communities is not supported for huawei")
            elif condition.operator is not ConditionOperator.HAS_ANY:
                raise NotImplementedError("Community operator %r not supported for huawei" % condition.operator)
            for comm_name in condition.value:
                yield "if-match community-filter", comm_name
            return
        if condition.field == MatchField.large_community:
            if condition.operator is ConditionOperator.HAS_ANY:
                if len(condition.value) > 1:
                    raise NotImplementedError("Multiple HAS_ANY values for large_community is not supported for huawei")
            elif condition.operator is not ConditionOperator.HAS:
                raise NotImplementedError("large_community operator %r not supported for huawei" % condition.operator)
            for comm_name in condition.value:
                yield "if-match large-community-filter", comm_name
            return
        if condition.field == MatchField.extcommunity_rt:
            if condition.operator is ConditionOperator.HAS:
                if len(condition.value) > 1:
                    raise NotImplementedError("Multiple HAS values for extcommunity_rt is not supported for huawei")
            elif condition.operator is not ConditionOperator.HAS_ANY:
                raise NotImplementedError("Extcommunity_rt operator %r not supported for huawei" % condition.operator)
            for comm_name in condition.value:
                if communities[comm_name].logic is CommunityLogic.AND:
                    yield "if-match extcommunity-filter", comm_name, "matches-all"
                else:
                    yield "if-match extcommunity-filter", comm_name
            return
        if condition.field == MatchField.extcommunity_soo:
            if condition.operator is ConditionOperator.HAS_ANY:
                if len(condition.value) > 1:
                    raise NotImplementedError("Multiple HAS_ANY for extcommunities_soo is not supported for huawei")
            elif condition.operator is not ConditionOperator.HAS:
                raise NotImplementedError("Extcommunity_soo operator %r not supported for huawei" % condition.operator)
            for comm_name in condition.value:
                yield "if-match extcommunity-list soo", comm_name
            return
        if condition.field == MatchField.rd:
            if len(condition.value) > 1:
                raise NotImplementedError("Multiple RD filters is not supported for huawei")
            rd_filter = rd_filters[condition.value[0]]
            yield "if-match rd-filter", str(rd_filter.number)
            return
        if condition.field == MatchField.ip_prefix:
            for name in condition.value.names:
                plist = name_generator.get_prefix(name, condition.value)
                yield "if-match", "ip-prefix", plist.name
            return
        if condition.field == MatchField.ipv6_prefix:
            for name in condition.value.names:
                plist = name_generator.get_prefix(name, condition.value)
                yield "if-match", "ipv6 address prefix-list", plist.name
            return
        if condition.field == MatchField.as_path_length:
            if condition.operator is ConditionOperator.EQ:
                yield "if-match", "as-path length", condition.value
            elif condition.operator is ConditionOperator.LE:
                yield "if-match", "as-path length less-equal", condition.value
            elif condition.operator is ConditionOperator.GE:
                yield "if-match", "as-path length greater-equal", condition.value
            elif condition.operator is ConditionOperator.BETWEEN_INCLUDED:
                yield "if-match", "as-path length greater-equal", condition.value[0], "less-equal", condition.value[1]
            else:
                raise NotImplementedError(
                    f"as_path_length operator {condition.operator} not supported for huawei",
                )
            return
        if condition.operator is not ConditionOperator.EQ:
            raise NotImplementedError(
                f"`{condition.field}` with operator {condition.operator} is not supported for huawei",
            )
        if condition.field not in HUAWEI_MATCH_COMMAND_MAP:
            raise NotImplementedError(f"Match using `{condition.field}` is not supported for huawei")
        cmd = HUAWEI_MATCH_COMMAND_MAP[condition.field]
        yield "if-match", cmd.format(option_value=condition.value)

    def _huawei_then_community(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            if action.value.added or action.value.removed:
                raise NotImplementedError(
                    "Cannot set community together with add/remove on huawei",
                )
            members = [m for name in action.value.replaced for m in communities[name].members]
            if members:
                yield "apply", "community", *members
            else:
                yield "apply", "community", "none"
        if action.value.added:
            members = [m for name in action.value.added for m in communities[name].members]
            yield "apply", "community", *members, "additive"
        for community_name in action.value.removed:
            yield "apply comm-filter", community_name, "delete"

    def _huawei_then_large_community(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            if action.value.added or action.value.removed:
                raise NotImplementedError(
                    "Cannot set large-community together with add/remove on huawei",
                )
            members = [m for name in action.value.replaced for m in communities[name].members]
            if members:
                yield "apply", "large-community", *members, "overwrite"
            else:
                yield "apply", "large-community", "none"
        if action.value.added:
            members = [m for name in action.value.added for m in communities[name].members]
            yield "apply", "large-community", *members, "additive"
        if action.value.removed:
            members = [m for name in action.value.removed for m in communities[name].members]
            yield "apply large-community", *members, "delete"

    def _huawei_then_extcommunity_rt(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            raise NotImplementedError("Extcommunity_rt replace is not supported for huawei")
        if action.value.added:
            members = [f"rt {m}" for name in action.value.added for m in communities[name].members]
            yield "apply", "extcommunity", *members, "additive"
        for community_name in action.value.removed:
            yield "apply extcommunity-filter rt", community_name, "delete"

    def _huawei_then_extcommunity_soo(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            raise NotImplementedError("Extcommunity_soo replace is not supported for huawei")
        if action.value.added:
            members = [f"rt {m}" for name in action.value.added for m in communities[name].members]
            yield "apply", "extcommunity", *members, "additive"
        if action.value.removed:
            raise NotImplementedError("Extcommunity_soo remove is not supported for huawei")

    def _huawei_render_ext_community_members(
            self, comm_type: CommunityType, members: list[str]
    ) -> Sequence[Sequence[str]]:
        if comm_type is CommunityType.SOO:
            return "soo", *members
        if comm_type is CommunityType.RT:
            return [f"rt {member}" for member in members]
        elif comm_type is CommunityType.LARGE:
            raise ValueError("Large community is not subtype of extcommunity")
        elif comm_type is CommunityType.BASIC:
            raise ValueError("Basic community is not subtype of extcommunity")
        else:
            raise NotImplementedError(f"Community type {comm_type} is not supported on huawei")

    def _huawei_then_extcommunity(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ):
        if action.value.replaced is not None:
            if action.value.added or action.value.removed:
                raise NotImplementedError(
                    "Cannot set extcommunity together with add/delete on huawei",
                )
            if not action.value.replaced:
                raise NotImplementedError(
                    "Cannot reset extcommunity on huawei",
                )

            members = group_community_members(communities, action.value.replaced)
            for community_type, replaced_members in members.items():
                if community_type is CommunityType.SOO:
                    raise NotImplementedError(
                        "Cannot set extcommunity soo on huawei",
                    )
                rendered_memebers = self._huawei_render_ext_community_members(community_type, replaced_members)
                yield "apply", "extcommunity", *rendered_memebers
        if action.value.added:
            members = group_community_members(communities, action.value.added)
            for community_type, added_members in members.items():
                rendered_memebers = self._huawei_render_ext_community_members(community_type, added_members)
                yield "apply", "extcommunity", *rendered_memebers, "additive"
        if action.value.removed:
            raise NotImplementedError("Cannot remove extcommunity on huawei")

    def _huawei_then_as_path(
            self,
            device: Any,
            action: SingleAction[AsPathActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.set is not None:
            if action.value.prepend:
                raise NotImplementedError(
                    "Cannot set as_path together with prepend on huawei",
                )
            if action.value.set:
                yield "apply", "as-path", *action.value.set, "overwrite"
            else:
                yield "apply", "as-path", "none overwrite"
        if action.value.prepend:
            yield "apply as-path", *action.value.prepend, "additive"
        if action.value.expand:
            raise RuntimeError("as_path.expand is not supported for huawei")
        if action.value.delete:
            for path_item in action.value.delete:
                yield "apply as-path", path_item, "delete"
        if action.value.expand_last_as:
            raise RuntimeError("as_path.expand_last_as is not supported for huawei")

    def _huawei_then(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[Any],
    ) -> Iterator[Sequence[str]]:
        if action.field == ThenField.community:
            yield from self._huawei_then_community(communities, device,
                                                   cast(SingleAction[CommunityActionValue], action))
            return
        if action.field == ThenField.large_community:
            yield from self._huawei_then_large_community(communities, device,
                                                         cast(SingleAction[CommunityActionValue], action))
            return
        if action.field == ThenField.extcommunity:
            yield from self._huawei_then_extcommunity(communities, device,
                                                      cast(SingleAction[CommunityActionValue], action))
            return
        if action.field == ThenField.extcommunity_rt:
            yield from self._huawei_then_extcommunity_rt(communities, device,
                                                         cast(SingleAction[CommunityActionValue], action))
            return
        if action.field == ThenField.extcommunity_soo:
            yield from self._huawei_then_extcommunity_soo(communities, device,
                                                          cast(SingleAction[CommunityActionValue], action))
            return
        if action.field == ThenField.metric:
            if action.type is ActionType.ADD:
                yield "apply", f"cost + {action.value}"
            elif action.type is ActionType.SET:
                yield "apply", f"cost {action.value}"
            else:
                raise NotImplementedError(f"Action type {action.type} for metric is not supported for huawei")
            return
        if action.field == ThenField.as_path:
            yield from self._huawei_then_as_path(device, cast(SingleAction[AsPathActionValue], action))
            return
        if action.field == ThenField.next_hop:
            next_hop_action_value = cast(NextHopActionValue, action.value)
            if next_hop_action_value.target == "self":
                yield "apply", "cost 1"
            elif next_hop_action_value.target == "discard":
                pass
            elif next_hop_action_value.target == "peer":
                pass
            elif next_hop_action_value.target == "ipv4_addr":
                yield "apply", f"ip-address next-hop {next_hop_action_value.addr}"
            elif next_hop_action_value.target == "ipv6_addr":
                yield "apply", f"ipv6 next-hop {next_hop_action_value.addr}"
            elif next_hop_action_value.target == "mapped_ipv4":
                yield "apply", f"ipv6 next-hop ::FFFF:{next_hop_action_value.addr}"
            else:
                raise RuntimeError(f"Next_hop target {next_hop_action_value.target} is not supported for huawei")

        if action.type is not ActionType.SET:
            raise NotImplementedError(f"Action type {action.type} for `{action.field}` is not supported for huawei")
        if action.field not in HUAWEI_THEN_COMMAND_MAP:
            raise NotImplementedError(f"Then action using `{action.field}` is not supported for huawei")
        cmd = HUAWEI_THEN_COMMAND_MAP[action.field]
        yield "apply", cmd.format(option_value=action.value)

    def _huawei_statement(
            self,
            communities: dict[str, CommunityList],
            rd_filters: dict[str, RDFilter],
            device: Any,
            policy: RoutingPolicy,
            statement: RoutingPolicyStatement,
            prefix_name_generator: PrefixListNameGenerator,
    ) -> Iterator[Sequence[str]]:
        if statement.number is None:
            raise RuntimeError(f"Statement number should not be empty on Huawei (found for policy: {policy.name})")
        with self.block(
                "route-policy", policy.name,
                HUAWEI_RESULT_MAP[statement.result],
                "node", statement.number
        ):
            for condition in statement.match:
                yield from self._huawei_match(device, condition, communities, rd_filters, prefix_name_generator)
            for action in statement.then:
                yield from self._huawei_then(communities, device, action)
            if statement.result is ResultType.NEXT:
                yield "goto next-node"

    def run_huawei(self, device):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)
        communities = {c.name: c for c in self.get_community_lists(device)}
        rd_filters = {f.name: f for f in self.get_rd_filters(device)}
        prefix_name_generator = PrefixListNameGenerator(prefix_lists, policies)

        for policy in self.get_policies(device):
            for statement in policy.statements:
                yield from self._huawei_statement(communities, rd_filters, device, policy, statement,
                                                  prefix_name_generator)

    # arista
    def acl_arista(self, device):
        return r"""
        route-map
            ~ %global=1
        """

    def _arista_match_community(
            self,
            device: Any,
            community_type: Literal["community", "extcommunity", "large-community"],
            community_names: Sequence[str],
    ) -> Iterator[Sequence[str]]:
        yield "match", community_type, *community_names

    def _arista_match(
            self,
            device: Any,
            condition: SingleCondition[Any],
            communities: dict[str, CommunityList],
            rd_filters: dict[str, RDFilter],
            name_generator: PrefixListNameGenerator,
    ) -> Iterator[Sequence[str]]:
        if condition.field == MatchField.community:
            if condition.operator is ConditionOperator.HAS_ANY:
                yield from self._arista_match_community(
                    device, "community", [mangle_united_community_list_name(condition.value)],
                )
            elif condition.operator is ConditionOperator.HAS:
                yield from self._arista_match_community(
                    device, "community", condition.value,
                )
            else:
                raise NotImplementedError(f"Community match operator {condition.field} is not supported on arista")
            return
        if condition.field == MatchField.large_community:
            if condition.operator is ConditionOperator.HAS_ANY:
                yield from self._arista_match_community(
                    device, "large-community", [mangle_united_community_list_name(condition.value)],
                )
            elif condition.operator is ConditionOperator.HAS:
                yield from self._arista_match_community(
                    device, "large-community", condition.value,
                )
            else:
                raise NotImplementedError(
                    f"Large-community match operator {condition.field} is not supported on arista")
            return
        if condition.field == MatchField.extcommunity_rt:
            if condition.operator is ConditionOperator.HAS_ANY:
                yield from self._arista_match_community(
                    device, "extcommunity", [mangle_united_community_list_name(condition.value)],
                )
            elif condition.operator is ConditionOperator.HAS:
                yield from self._arista_match_community(
                    device, "extcommunity", condition.value,
                )
            else:
                raise NotImplementedError(f"Community match operator {condition.field} is not supported on arista")
            return
        if condition.field == MatchField.extcommunity_soo:
            if condition.operator is ConditionOperator.HAS_ANY:
                yield from self._arista_match_community(
                    device, "extcommunity", [mangle_united_community_list_name(condition.value)],
                )
            elif condition.operator is ConditionOperator.HAS:
                yield from self._arista_match_community(
                    device, "extcommunity", condition.value,
                )
            else:
                raise NotImplementedError(f"Extcommunity match operator {condition.field} is not supported on arista")
            return
        if condition.field == MatchField.ip_prefix:
            for name in condition.value.names:
                plist = name_generator.get_prefix(name, condition.value)
                yield "match", "ip address prefix-list", plist.name
            return
        if condition.field == MatchField.ipv6_prefix:
            for name in condition.value.names:
                plist = name_generator.get_prefix(name, condition.value)
                yield "match", "ipv6 address prefix-list", plist.name
            return
        if condition.field == MatchField.as_path_length:
            if condition.operator is ConditionOperator.EQ:
                yield "match", "as-path length =", condition.value
            elif condition.operator is ConditionOperator.LE:
                yield "match", "as-path length <=", condition.value
            elif condition.operator is ConditionOperator.GE:
                yield "match", "as-path length >=", condition.value
            elif condition.operator is ConditionOperator.BETWEEN_INCLUDED:
                yield "match", "as-path length >=", condition.value[0]
                yield "match", "as-path length <=", condition.value[1]
            else:
                raise NotImplementedError(
                    f"as_path_length operator {condition.operator} not supported for arista",
                )
            return
        if condition.operator is not ConditionOperator.EQ:
            raise NotImplementedError(
                f"`{condition.field}` with operator {condition.operator} is not supported for arista",
            )
        if condition.field not in ARISTA_MATCH_COMMAND_MAP:
            raise NotImplementedError(f"Match using `{condition.field}` is not supported for arista")
        cmd = ARISTA_MATCH_COMMAND_MAP[condition.field]
        yield "match", cmd.format(option_value=condition.value)

    def _arista_then_community(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            if action.value.added or action.value.removed:
                raise NotImplementedError(
                    "Cannot set community together with add/remove on arista",
                )

            if action.value.replaced:
                yield "set", "community community-list", *action.value.replaced
            else:
                yield "set", "community", "none"
        if action.value.added:
            yield "set", "community community-list", *action.value.added, "additive"
        if action.value.removed:
            members = [
                arista_well_known_community(member)
                for community_name in action.value.removed
                for member in communities[community_name].members
            ]
            yield "set community", *members, "delete"

    def _arista_then_large_community(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            if action.value.added or action.value.removed:
                raise NotImplementedError(
                    "Cannot set large-community together with add/remove on arista",
                )

            if not action.value.replaced:
                yield "set", "large-community", "none"
            first = True
            for community_name in action.value.replaced:
                if first:
                    yield "set", "large-community large-community-list", community_name
                    first = False
                else:
                    yield "set", "large-community large-community-list", community_name, "additive"
        if action.value.added:
            yield "set", "large-community large-community-list", *action.value.added, "additive"
        if action.value.removed:
            yield "set large-community large-community-list", *action.value.removed, "delete"

    def _arista_then_extcommunity_rt(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            raise NotImplementedError("Extcommunity_rt replace is not supported for arista")
        if action.value.added:
            members = [
                f"rt {member}"
                for community_name in action.value.removed
                for member in communities[community_name].members
            ]
            yield "set", "extcommunity", *members, "additive"
        if action.value.removed:
            members = [
                f"rt {member}"
                for community_name in action.value.removed
                for member in communities[community_name].members
            ]
            yield "set extcommunity", *members, "delete"

    def _arista_then_extcommunity_soo(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            raise NotImplementedError("Extcommunity_soo replace is not supported for arista")
        if action.value.added:
            members = [
                f"soo {member}"
                for community_name in action.value.removed
                for member in communities[community_name].members
            ]
            yield "set", "extcommunity", *members, "additive"
        if action.value.removed:
            members = [
                f"soo {member}"
                for community_name in action.value.removed
                for member in communities[community_name].members
            ]
            yield "set", "extcommunity", *members, "delete"

    def _arista_extcommunity_type_str(self, comm_type: CommunityType) -> str:
        if comm_type is CommunityType.SOO:
            return "soo"
        elif comm_type is CommunityType.RT:
            return "rt"
        elif comm_type is CommunityType.LARGE:
            raise ValueError("Large community is not subtype of extcommunity")
        elif comm_type is CommunityType.BASIC:
            raise ValueError("Basic community is not subtype of extcommunity")
        else:
            raise NotImplementedError(f"Community type {comm_type} is not supported on arista")

    def _arista_render_ext_community_members(
            self, all_communities: dict[str, CommunityList], communities: list[str],
    ) -> Iterator[str]:
        for community_name in communities:
            community = all_communities[community_name]
            comm_type = self._arista_extcommunity_type_str(community.type)
            for member in community.members:
                yield f"{comm_type} {member}"

    def _arista_then_extcommunity(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ):
        if action.value.replaced is not None:
            if action.value.added or action.value.removed:
                raise NotImplementedError(
                    "Cannot set extcommunity together with add/delete on arista",
                )
            if not action.value.replaced:
                yield "set", "extcommunity", "none"
                return
            members = list(self._arista_render_ext_community_members(communities, action.value.replaced))
            yield "set extcommunity", *members
            return
        if action.value.added:
            members = list(self._arista_render_ext_community_members(communities, action.value.added))
            yield "set extcommunity", *members, "additive"
        if action.value.removed:
            members = list(self._arista_render_ext_community_members(communities, action.value.removed))
            yield "set extcommunity", *members, "delete"

    def _arista_then_as_path(
            self,
            device: Any,
            action: SingleAction[AsPathActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.set is not None:
            if action.value.prepend:
                raise NotImplementedError(
                    "Cannot set as_path together with prepend on arista",
                )
            if not action.value.set:
                yield "set", "as-path match all replacement", "none"
            else:
                yield "set", "as-path match all replacement", *action.value.set

        if action.value.prepend:
            yield "set", "as-path prepend", *action.value.prepend
        if action.value.expand_last_as:
            yield "set", "as-path prepend last-as", action.value.expand_last_as

        if action.value.expand:
            raise RuntimeError("as_path.expand is not supported for arista")
        if action.value.delete:
            raise RuntimeError("as_path.delete is not supported for arista")

    def _arista_then(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[Any],
    ) -> Iterator[Sequence[str]]:
        if action.field == ThenField.community:
            yield from self._arista_then_community(
                communities, device, cast(SingleAction[CommunityActionValue], action),
            )
            return
        if action.field == ThenField.large_community:
            yield from self._arista_then_large_community(
                communities, device, cast(SingleAction[CommunityActionValue], action),
            )
            return
        if action.field == ThenField.extcommunity:
            yield from self._arista_then_extcommunity(communities, device, action)
            return
        if action.field == ThenField.extcommunity_rt:
            yield from self._arista_then_extcommunity_rt(
                communities, device, cast(SingleAction[CommunityActionValue], action),
            )
            return
        if action.field == ThenField.extcommunity_soo:
            yield from self._arista_then_extcommunity_soo(
                communities, device, cast(SingleAction[CommunityActionValue], action),
            )
            return
        if action.field == ThenField.metric:
            if action.type is ActionType.ADD:
                yield "set", f"metric + {action.value}"
            elif action.type is ActionType.REMOVE:
                yield "set", f"metric - {action.value}"
            elif action.type is ActionType.SET:
                yield "set", f"metric {action.value}"
            else:
                raise NotImplementedError(f"Action type {action.type} for metric is not supported for arista")
            return
        if action.field == ThenField.as_path:
            yield from self._arista_then_as_path(device, cast(SingleAction[AsPathActionValue], action))
            return
        if action.field == ThenField.next_hop:
            next_hop_action_value = cast(NextHopActionValue, action.value)
            if next_hop_action_value.target == "self":
                yield "set", "cost 1"  # TODO?
            elif next_hop_action_value.target == "discard":
                pass
            elif next_hop_action_value.target == "peer":
                pass
            elif next_hop_action_value.target == "ipv4_addr":
                yield "set", f"ip next-hop {next_hop_action_value.addr}"
            elif next_hop_action_value.target == "ipv6_addr":
                yield "set", f"ipv6 next-hop {next_hop_action_value.addr}"
            elif next_hop_action_value.target == "mapped_ipv4":
                yield "set", f"ipv6 next-hop ::FFFF:{next_hop_action_value.addr}"
            else:
                raise RuntimeError(f"Next_hop target {next_hop_action_value.target} is not supported for arista")
            return
        if action.type is not ActionType.SET:
            raise NotImplementedError(f"Action type {action.type} for `{action.field}` is not supported for arista")
        if action.field not in ARISTA_THEN_COMMAND_MAP:
            raise NotImplementedError(f"Then action using `{action.field}` is not supported for arista")
        cmd = ARISTA_THEN_COMMAND_MAP[action.field]
        yield "set", cmd.format(option_value=action.value)

    def _arista_statement(
            self,
            communities: dict[str, CommunityList],
            rd_filters: dict[str, RDFilter],
            device: Any,
            policy: RoutingPolicy,
            statement: RoutingPolicyStatement,
            prefix_name_generator: PrefixListNameGenerator,
    ) -> Iterator[Sequence[str]]:
        if statement.number is None:
            raise RuntimeError(f"Statement number should not be empty on Arista (found for policy: {policy.name})")
        with self.block(
                "route-map",
                policy.name,
                ARISTA_RESULT_MAP[statement.result],
                statement.number,
        ):
            for condition in statement.match:
                yield from self._arista_match(device, condition, communities, rd_filters, prefix_name_generator)
            for action in statement.then:
                yield from self._arista_then(communities, device, action)
            if statement.result is ResultType.NEXT:
                yield "continue"

    def run_arista(self, device):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)
        prefix_name_generator = PrefixListNameGenerator(prefix_lists, policies)
        communities = {c.name: c for c in self.get_community_lists(device)}
        rd_filters = {f.name: f for f in self.get_rd_filters(device)}

        for policy in policies:
            for statement in policy.statements:
                yield from self._arista_statement(
                    communities, rd_filters, device, policy, statement, prefix_name_generator,
                )

    # Cisco IOS XR
    def acl_iosxr(self, device):
        return r"""
        route-policy *
            ~ %global=1
        """

    def _iosxr_match_community(
            self,
            community_type: Literal["community", "extcommunity rt", "extcommunity soo"],
            community: CommunityList,
    ) -> Iterator[Sequence[str]]:
        if community.logic is CommunityLogic.AND:
            yield community_type, "matches-every", community.name
        elif community.logic is CommunityLogic.OR:
            yield community_type, "matches-any", community.name
        else:
            raise ValueError(f"Unknown community logic {community.logic}")

    def _iosxr_match_communities(
            self,
            operator: ConditionOperator,
            community_type: Literal["community", "extcommunity rt", "extcommunity soo"],
            community_names: list[str],
            communities: dict[str, CommunityList],
    ) -> Iterator[Sequence[str]]:
        if operator == ConditionOperator.HAS_ANY:
            yield self._iosxr_or_matches(
                self._iosxr_match_community(community_type, communities[name])
                for name in community_names
            ),
        elif operator == ConditionOperator.HAS:
            for name in community_names:
                yield from self._iosxr_match_community(community_type, communities[name])
        else:
            raise NotImplementedError(f"Operator {operator} is not supported for {community_type} on Cisco IOS XR")

    def _iosxr_match_rd(
            self,
            condition: SingleCondition[Sequence[str]],
            rd_filters: dict[str, RDFilter],
    ) -> Iterator[Sequence[str]]:
        if condition.operator == ConditionOperator.HAS_ANY:
            if len(condition.value) == 1:
                yield "rd", "in", condition.value[0]
                return
            conds = " or ".join(f"rd in {name}" for name in condition.value)
            yield f"({conds})",
            return
        elif condition.operator == ConditionOperator.HAS:
            for name in condition.value:
                yield "rd", "in", name
        else:
            raise NotImplementedError(f"Operator {condition.operator} is not supported for {condition.field} on Cisco IOS XR")

    def _iosxr_match_prefix_lists(
            self,
            condition: SingleCondition[PrefixMatchValue],
            name_generator: PrefixListNameGenerator
    ) -> Iterator[Sequence[str]]:
        matches = []
        for name in condition.value.names:
            plist = name_generator.get_prefix(name, condition.value)
            matches.append(("destination in", plist.name))
        if len(matches) == 1:
            yield matches[0]
        else:
            yield self._iosxr_or_matches([matches]),

    def _iosxr_match_local_pref(self, condition: SingleCondition) -> Iterator[Sequence[str]]:
        if condition.operator is ConditionOperator.EQ:
            yield "local-preference", "eq", str(condition.value)
        elif condition.operator is ConditionOperator.LE:
            yield "local-preference", "le", str(condition.value)
        elif condition.operator is ConditionOperator.GE:
            yield "local-preference", "ge", str(condition.value)
        elif condition.operator is ConditionOperator.BETWEEN_INCLUDED:
            yield "local-preference", "ge", str(condition.value[0])
            yield "local-preference", "le", str(condition.value[1])
        else:
            raise NotImplementedError(
                f"Operator {condition.operator} is not supported for {condition.field} on Cisco IOS XR",
            )

    def _iosxr_match_as_path_length(self, condition: SingleCondition) -> Iterator[Sequence[str]]:
        if condition.operator is ConditionOperator.EQ:
            yield "as-path length", "eq", str(condition.value)
        elif condition.operator is ConditionOperator.LE:
            yield "as-path length", "le", str(condition.value)
        elif condition.operator is ConditionOperator.GE:
            yield "as-path length", "ge", str(condition.value)
        elif condition.operator is ConditionOperator.BETWEEN_INCLUDED:
            yield "as-path length", "ge", str(condition.value[0])
            yield "as-path length", "le", str(condition.value[1])
        else:
            raise NotImplementedError(
                f"Operator {condition.operator} is not supported for {condition.field} on Cisco IOS XR",
            )

    def _iosxr_and_matches(self, conditions: Iterable[Iterable[Sequence[str]]]) -> str:
        return " and ".join(" ".join(c) for seq in conditions for c in seq)

    def _iosxr_or_matches(self, conditions: Iterable[Iterable[Sequence[str]]]) -> str:
        return "(" + " or ".join(" ".join(c) for seq in conditions for c in seq) + ")"

    def _iosxr_match(
            self,
            device: Any,
            condition: SingleCondition[Any],
            communities: dict[str, CommunityList],
            rd_filters: dict[str, RDFilter],
            name_generator: PrefixListNameGenerator,
    ) -> Iterator[Sequence[str]]:
        if condition.field == MatchField.community:
            yield from self._iosxr_match_communities(condition.operator, "community", condition.value, communities)
            return
        elif condition.field == MatchField.extcommunity_rt:
            yield from self._iosxr_match_communities(condition.operator, "extcommunity rt", condition.value, communities)
            return
        elif condition.field == MatchField.extcommunity_soo:
            yield from self._iosxr_match_communities(condition.operator, "extcommunity soo", condition.value, communities)
            return
        elif condition.field == MatchField.local_pref:
            yield from self._iosxr_match_local_pref(condition)
            return
        elif condition.field == MatchField.as_path_length:
            yield from self._iosxr_match_as_path_length(condition)
            return
        elif condition.field == MatchField.rd:
            yield from self._iosxr_match_rd(condition, rd_filters)
            return
        elif condition.field == MatchField.ip_prefix:
            yield from self._iosxr_match_prefix_lists(condition, name_generator)
            return
        elif condition.field == MatchField.ipv6_prefix:
            yield from self._iosxr_match_prefix_lists(condition, name_generator)
            return

        if condition.operator is not ConditionOperator.EQ:
            raise NotImplementedError(
                f"`{condition.field}` with operator {condition.operator} is not supported for Cisco IOS XR",
            )
        if condition.field not in IOSXR_MATCH_COMMAND_MAP:
            raise NotImplementedError(f"Match using `{condition.field}` is not supported for Cisco IOS XR")
        yield IOSXR_MATCH_COMMAND_MAP[condition.field].format(option_value=condition.value),

    def _iosxr_then_as_path(
            self,
            device: Any,
            action: SingleAction[AsPathActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.set is not None:
            raise RuntimeError("as_path.set is not supported for Cisco IOS XR")
        if action.value.prepend:
            for asn in action.value.prepend:
                yield "prepend as-path", asn
        if action.value.expand:
            raise RuntimeError("as_path.expand is not supported for Cisco IOS XR")
        if action.value.delete:
            raise RuntimeError("as_path.delete is not supported for Cisco IOS XR")
        if action.value.expand_last_as:
            raise RuntimeError("as_path.expand_last_as is not supported for Cisco IOS XR")

    def _iosxr_then_next_hop(
            self,
            device: Any,
            action: SingleAction[NextHopActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.target == "self":
            yield "set", "next-hop", "self"
        elif action.value.target == "discard":
            yield "set", "next-hop", "discard"
        elif action.value.target == "peer":
            pass
        elif action.value.target == "ipv4_addr":
            yield "set", "next-hop", action.value.addr
        elif action.value.target == "ipv6_addr":
            yield "set", "next-hop", action.value.addr.lower()
        elif action.value.target == "mapped_ipv4":
            yield "set", "next-hop", f"::ffff:{action.value.addr}"
        else:
            raise RuntimeError(f"Next_hop target {action.value.target} is not supported for Cisco IOS XR")

    def _iosxr_then_metric(
            self,
            device: Any,
            action: SingleAction[NextHopActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.type is ActionType.ADD:
            yield "set", f"med +{action.value}"
        elif action.type is ActionType.REMOVE:
            yield "set", f"med -{action.value}"
        elif action.type is ActionType.SET:
            yield "set", f"med {action.value}"
        else:
            raise NotImplementedError(f"Action type {action.type} for metric is not supported for Cisco IOS XR")

    def _iosxr_then_community(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        added = []
        if action.value.replaced is not None:
            yield "delete", "community", "all"
            added.extend(action.value.replaced)
        added.extend(action.value.added)
        for name in added:
            yield "set", "community", name, "additive"
        for community_name in action.value.removed:
            yield "delete", "community", "in", community_name

    def _iosxr_community_type_str(self, comm_type: CommunityType) -> str:
        if comm_type is CommunityType.RT:
            return "rt"
        elif comm_type is CommunityType.LARGE:
            raise ValueError("Large community is not subtype of extcommunity")
        elif comm_type is CommunityType.BASIC:
            raise ValueError("Basic community is not subtype of extcommunity")
        else:
            raise NotImplementedError(f"Community type {comm_type} is not supported on Cisco IOS XR")

    def _iosxr_then_extcommunity(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        added = []
        if action.value.replaced is not None:
            yield "delete", "extcommunity", "rt", "all"
            added.extend(action.value.replaced)

        added.extend(action.value.added)
        for member_name in added:
            typename = self._iosxr_community_type_str(communities[member_name].type)
            yield "set", "extcommunity", typename, member_name, "additive"
        for member_name in action.value.removed:
            typename = self._iosxr_community_type_str(communities[member_name].type)
            yield "delete", "extcommunity", typename, "in", member_name

    def _iosxr_then(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[Any],
    ) -> Iterator[Sequence[str]]:
        if action.field == ThenField.community:
            yield from self._iosxr_then_community(communities, device, action)
            return
        if action.field == ThenField.extcommunity:
            yield from self._iosxr_then_extcommunity(communities, device, action)
            return
        if action.field == ThenField.metric:
            yield from self._iosxr_then_metric(device, action)
            return
        if action.field == ThenField.as_path:
            yield from self._iosxr_then_as_path(device, cast(SingleAction[AsPathActionValue], action))
            return
        if action.field == ThenField.next_hop:
            yield from self._iosxr_then_next_hop(device, cast(SingleAction[NextHopActionValue], action))
            return
        if action.type is not ActionType.SET:
            raise NotImplementedError(f"Action type {action.type} for `{action.field}` is not supported for Cisco IOS XR")
        if action.field not in IOSXR_THEN_COMMAND_MAP:
            raise NotImplementedError(f"Then action using `{action.field}` is not supported for Cisco IOS XR")
        cmd = IOSXR_THEN_COMMAND_MAP[action.field]
        yield "set", cmd.format(option_value=action.value)

    def _iosxr_statement(
            self,
            communities: dict[str, CommunityList],
            rd_filters: dict[str, RDFilter],
            device: Any,
            policy: RoutingPolicy,
            statement: RoutingPolicyStatement,
            prefix_name_generator: PrefixListNameGenerator,
    ) -> Iterator[Sequence[str]]:
        if statement.match:
            condition_expr = self._iosxr_and_matches(
                self._iosxr_match(device, condition, communities, rd_filters, prefix_name_generator)
                for condition in statement.match
            )
            with self.block(
                    "if", condition_expr, "then",
            ):
                for action in statement.then:
                    yield from self._iosxr_then(communities, device, action)
                yield IOSXR_RESULT_MAP[statement.result],
        else:
            for action in statement.then:
                yield from self._iosxr_then(communities, device, action)
            yield IOSXR_RESULT_MAP[statement.result],

    def run_iosxr(self, device):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)
        prefix_name_generator = PrefixListNameGenerator(prefix_lists, policies)
        communities = {c.name: c for c in self.get_community_lists(device)}
        rd_filters = {f.name: f for f in self.get_rd_filters(device)}

        for policy in policies:
            with self.block(
                    "route-policy", policy.name,
            ):
                for statement in policy.statements:
                    yield from self._iosxr_statement(
                        communities, rd_filters, device, policy, statement, prefix_name_generator,
                    )

    # Juniper
    def acl_juniper(self, device):
        return r"""
        policy-options       %cant_delete
            policy-statement
                ~            %global
        """

    def _juniper_match_communities(
            self,
            section: Literal["", "from"],
            conditions: list[SingleCondition],
    ) -> Iterator[Sequence[str]]:
        names: list[str] = [name for cond in conditions for name in cond.value]
        operators = {x.operator for x in conditions}
        if len(names) > 1 and operators != {ConditionOperator.HAS_ANY}:
            raise NotImplementedError(
                f"Multiple community match [{' '.join(names)}] without has_any is not supported for Juniper",
            )
        yield section, "community", self._juniper_list_bracket(names)

    def _juniper_match_prefix_lists(
            self,
            section: Literal["", "from"],
            conditions: list[SingleCondition[PrefixMatchValue]],
            name_generator: JuniperPrefixListNameGenerator,
    ) -> Iterator[Sequence[str]]:
        operators = {x.operator for x in conditions}
        supported = {ConditionOperator.HAS_ANY}
        not_supported = operators - supported
        if len(conditions) > 1 and not_supported:
            raise NotImplementedError(
                f"Multiple prefix match with ops {not_supported} is not supported for Juniper",
            )
        for cond in conditions:
            for name in cond.value.names:
                prefix_list = name_generator.get_prefix(name, cond.value)
                plist_type = name_generator.get_type(name, cond.value)
                flavour = name_generator.get_plist_flavour(prefix_list)
                if plist_type == "prefix-list" and flavour == "simple":
                    yield section, "prefix-list", prefix_list.name
                elif plist_type == "prefix-list" and flavour == "orlonger":
                    yield section, "prefix-list-filter", prefix_list.name, "orlonger"
                elif plist_type == "route-filter":
                    yield section, "route-filter-list", prefix_list.name
                else:
                    raise NotImplementedError(
                        f"Prefix list {prefix_list.name} type {plist_type} flavour {flavour} is not supported for Juniper",
                    )

    def _juniper_match_as_path_length(
            self,
            section: Literal["", "from"],
            conditions: list[SingleCondition],
    ) -> Iterator[Sequence[str]]:
        for condition in conditions:
            if condition.operator is ConditionOperator.EQ:
                yield section, "as-path-calc-length", str(condition.value), "equal"
            elif condition.operator is ConditionOperator.LE:
                yield section, "as-path-calc-length", str(condition.value), "orlower"
            elif condition.operator is ConditionOperator.GE:
                yield section, "as-path-calc-length", str(condition.value), "orhigher"
            elif condition.operator is ConditionOperator.BETWEEN_INCLUDED:
                yield section, "as-path-calc-length", str(condition.value[0]), "orhigher"
                yield section, "as-path-calc-length", str(condition.value[1]), "orlower"
            else:
                raise NotImplementedError(
                    f"Operator {condition.operator} is not supported for {condition.field} on Juniper",
                )

    def _juniper_match_rd_filter(
            self,
            section: Literal["", "from"],
            conditions: list[SingleCondition[Sequence[str]]],
    ) -> Iterator[Sequence[str]]:
        names = [x for c in conditions for x in c.value]
        operators = {x.operator for x in conditions}
        supported = {ConditionOperator.HAS_ANY}
        not_supported = operators - supported
        if len(names) > 1 and not_supported:
            raise NotImplementedError(
                f"Multiple rd_filter matches with ops {not_supported} is not supported for Juniper",
            )
        yield section, "route-distinguisher", self._juniper_list_bracket(names)

    def _juniper_match_community_fields(self) -> set[MatchField]:
        return {
            MatchField.community,
            MatchField.extcommunity_rt,
            MatchField.extcommunity_soo,
            MatchField.large_community,
        }

    def _juniper_match_prefix_fields(self) -> set[MatchField]:
        return {
            MatchField.ip_prefix,
            MatchField.ipv6_prefix,
        }

    def _juniper_is_match_inlined(self, conditions: AndCondition) -> bool:
        used_fields = {x.field for x in conditions}
        used_prefix_fields = used_fields & self._juniper_match_prefix_fields()
        used_community_fields = used_fields & self._juniper_match_community_fields()

        # prefix-list match is never inlined
        if used_prefix_fields:
            return False

        # as-path-calc-length is never inlined
        if MatchField.as_path_length in used_fields:
            return False

        # only community matches and nothing more
        if used_community_fields and used_fields == used_community_fields:
            return True

        # inline when empty or just one match
        if len(used_fields) <= 1:
            return True
        return False

    def _juniper_match(
            self,
            policy: RoutingPolicy,
            section: Literal["", "from"],
            conditions: AndCondition,
            prefix_name_generator: JuniperPrefixListNameGenerator,
    ) -> Iterator[Sequence[str]]:
        community_fields = self._juniper_match_community_fields()
        prefix_fields = self._juniper_match_prefix_fields()
        community_conditions: list[SingleCondition] = []
        prefix_conditions: list[SingleCondition] = []
        simple_conditions: list[SingleCondition] = []
        as_path_length_conditions: list[SingleCondition] = []
        rd_filter_conditions: list[SingleCondition] = []
        for condition in conditions:
            if condition.field in community_fields:
                community_conditions.append(condition)
            elif condition.field in prefix_fields:
                prefix_conditions.append(condition)
            elif condition.field == MatchField.as_path_length:
                as_path_length_conditions.append(condition)
            elif condition.field == MatchField.rd:
                rd_filter_conditions.append(condition)
            else:
                simple_conditions.append(condition)

        if community_conditions:
            yield from self._juniper_match_communities(section, community_conditions)
        if prefix_conditions:
            yield from self._juniper_match_prefix_lists(section, prefix_conditions, prefix_name_generator)
        if as_path_length_conditions:
            yield from self._juniper_match_as_path_length(section, as_path_length_conditions)
        if rd_filter_conditions:
            yield from self._juniper_match_rd_filter(section, rd_filter_conditions)

        for condition in simple_conditions:
            if condition.operator is not ConditionOperator.EQ:
                raise NotImplementedError(
                    f"`{condition.field}` with operator {condition.operator} in {policy.name} is not supported for Juniper",
                )
            if condition.field not in JUNIPER_MATCH_COMMAND_MAP:
                raise NotImplementedError(f"Match using `{condition.field}` in {policy.name} is not supported for Juniper")
            yield section, JUNIPER_MATCH_COMMAND_MAP[condition.field].format(option_value=condition.value)

    def _juniper_then_community(
            self,
            section: Literal["", "then"],
            actions: list[SingleAction[CommunityActionValue]]
        ):
        # juniper community ops are ORDERED
        # since data model does not support it
        # we use the order that makes sense: delete, set, add
        for single_action in actions:
            action = single_action.value
            for name in action.removed:
                yield section, "community", "delete", name

            if action.replaced is not None:
                if not action.replaced:
                    raise NotImplementedError("Empty community.set() is not supported for Juniper")
                for name in action.replaced:
                    yield section, "community", "set", name

            for name in action.added:
                yield section, "community", "add", name

    def _juniper_then_next_hop(
            self,
            section: Literal["", "then"],
            actions: list[SingleAction[NextHopActionValue]],
        ):
        if len(actions) > 1:
            raise NotImplementedError("Only single next-hop action is supported for Juniper")

        action = actions[0]
        if action.value.target == "self":
            yield section, "next-hop", "self"
        elif action.value.target == "discard":
            yield section, "next-hop", "discard"
        elif action.value.target == "peer":
            yield section, "next-hop", "peer-address"
        elif action.value.target == "ipv4_addr":
            yield section, "next-hop", action.value.addr
        elif action.value.target == "ipv6_addr":
            yield section, "next-hop", action.value.addr.lower()
        elif action.value.target == "mapped_ipv4":
            yield section, "next-hop", f"::ffff:{action.value.addr}"
        else:
            raise NotImplementedError(f"Next_hop target {action.value.target} is not supported for Juniper")

    def _juniper_list_quote(self, items: list[str]) -> str:
        joined = " ".join(items)
        if len(items) > 1:
            joined = f'"{joined}"'
        return joined

    def _juniper_list_bracket(self, items: list[str]) -> str:
        joined = " ".join(items)
        if len(items) > 1:
            joined = f"[ {joined} ]"
        return joined

    def _juniper_then_as_path(
            self,
            section: Literal["", "then"],
            actions: list[SingleAction[AsPathActionValue]],
        ):
        if len(actions) > 1:
            raise NotImplementedError("Only single next-hop action is supported for Juniper")

        action = actions[0]
        if action.value.expand and action.value.expand_last_as:
            raise NotImplementedError("Setting both `as_path.expand` and `as_path.expand_last_as` is not supported for Juniper")

        if action.value.prepend:
            yield section, "as-path-prepend", self._juniper_list_quote(action.value.prepend)
        if action.value.expand:
            yield section, "as-path-expand", self._juniper_list_quote(action.value.expand)
        if action.value.expand_last_as:
            yield section, "as-path-expand last-as count", action.value.expand_last_as
        if action.value.set is not None:
            raise RuntimeError("as_path.set is not supported for Juniper")
        if action.value.delete:
            raise RuntimeError("as_path.delete is not supported for Juniper")

    def _juniper_is_then_inlined(self, action: Action) -> bool:
        used_fields = {x.field for x in action}
        # inline when no actions permormed
        if not used_fields:
            return True
        return False

    def _juniper_then(
            self,
            policy: RoutingPolicy,
            section: Literal["", "then"],
            actions: Action,
    ) -> Iterator[Sequence[str]]:
        community_actions: list[SingleAction] = []
        next_hop_actions: list[SingleAction] = []
        as_path_actions: list[SingleAction] = []
        simple_actions: list[SingleAction] = []
        for action in actions:
            if action.field == ThenField.community:
                community_actions.append(action)
            elif action.field == ThenField.extcommunity:
                community_actions.append(action)
            elif action.field == ThenField.extcommunity_rt:
                community_actions.append(action)
            elif action.field == ThenField.extcommunity_soo:
                community_actions.append(action)
            elif action.field == ThenField.large_community:
                community_actions.append(action)
            elif action.field == ThenField.next_hop:
                next_hop_actions.append(action)
            elif action.field == ThenField.as_path:
                as_path_actions.append(action)
            else:
                simple_actions.append(action)

        if community_actions:
            yield from self._juniper_then_community(section, community_actions)
        if next_hop_actions:
            yield from self._juniper_then_next_hop(section, next_hop_actions)
        if as_path_actions:
            yield from self._juniper_then_as_path(section, as_path_actions)

        for action in simple_actions:
            if action.type not in {ActionType.SET}:
                raise NotImplementedError(f"Action type {action.type} for `{action.field}` in {policy.name} is not supported for Juniper")
            if action.field not in JUNIPER_THEN_COMMAND_MAP:
                raise NotImplementedError(f"Then action using `{action.field}` in {policy.name} is not supported for Juniper")
            yield section, JUNIPER_THEN_COMMAND_MAP[action.field].format(option_value=action.value)

    def _juniper_statements(
            self,
            device: Any,
            policy: RoutingPolicy,
            prefix_name_generator: JuniperPrefixListNameGenerator,
    ) -> Iterator[Sequence[str]]:
        term_number = 0
        for statement in policy.statements:
            if statement.number is not None:
                term_number = statement.number
            term_name = statement.name
            if not term_name:
                term_name = f"{policy.name}_{term_number}"
            term_number += 1

            with self.block("term", term_name):
                # see test_juniper_inline
                match_inlined = self._juniper_is_match_inlined(statement.match)
                then_inlined = self._juniper_is_then_inlined(statement.then)
                match_section: Literal["", "from"] = "from" if match_inlined else ""
                then_section: Literal["", "then"] = "then" if then_inlined else ""

                if statement.match:
                    with self.block_if("from", condition=not match_inlined):
                        yield from self._juniper_match(policy, match_section, statement.match, prefix_name_generator)

                if statement.then:
                    with self.block_if("then", condition=not then_inlined):
                        yield from self._juniper_then(policy, then_section, statement.then)

                with self.block_if("then", condition=not then_inlined):
                    yield then_section, JUNIPER_RESULT_MAP[statement.result]

    def run_juniper(self, device):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)
        prefix_name_generator = JuniperPrefixListNameGenerator(prefix_lists, policies)

        for policy in policies:
            with self.block("policy-options"):
                with self.block("policy-statement", policy.name):
                    yield from self._juniper_statements(
                        device, policy, prefix_name_generator,
                    )



from __future__ import annotations

import pkgutil
import re
import types
from typing import FrozenSet, Iterable, List, Optional, Set, Union

from annet.lib import flatten, mako_render

from .base import BaseGenerator, _filter_str
from .exceptions import NotSupportedDevice


class Entire(BaseGenerator):
    TYPE = "ENTIRE"
    TAGS: List[str] = []
    REQUIRED_PACKAGES: FrozenSet[str] = frozenset()
    ENSURE_END_NEWLINE = False

    def __init__(self, storage):
        self.storage = storage
        # между генераторами для одного и того же path - выбирается тот что больше
        if not hasattr(self, "prio"):
            self.prio = 100
        self.__device = None

    def supports_device(self, device):
        return bool(self.path(device))

    def run(self, device) -> Union[None, str, Iterable[Union[str, tuple]]]:
        raise NotImplementedError

    def reload(self, device) -> Optional[str]:  # pylint: disable=unused-argument
        return

    def get_reload_cmds(self, device) -> str:
        ret = self.reload(device) or ""
        path = self.path(device)
        if path and device.hw.PC and device.hw.soft.startswith(
            ("Cumulus", "SwitchDev", "SONiC"),
        ):
            parts = []
            if ret:
                parts.append(ret)
            parts.append("/usr/bin/etckeeper commitreload %s" % path)
            return "\n".join(parts)
        return ret

    def path(self, device) -> Optional[str]:
        raise NotImplementedError("Required PATH for ENTIRE generator")

    # pylint: disable=unused-argument
    def is_safe(self, device) -> bool:
        """Output gen results when --acl-safe flag is used"""
        return False

    def read(self, path) -> str:
        return pkgutil.get_data(__name__, path).decode()

    def mako(self, text, **kwargs) -> str:
        return mako_render(text, dedent=True, device=self.__device, **kwargs)

    # =====

    @classmethod
    def get_aliases(cls) -> Set[str]:
        return {cls.__name__, *cls.TAGS}

    def __call__(self, device):
        self.__device = device
        parts = []
        run_res = self.run(device)
        if isinstance(run_res, str):
            run_res = (run_res,)
        if run_res is None or not isinstance(run_res, (tuple, types.GeneratorType)):
            raise Exception("generator %s returns %s" % (
                self.__class__.__name__, type(run_res)))
        for text in run_res:
            if isinstance(text, tuple):
                text = " ".join(map(_filter_str, flatten(text)))
            assert re.search(r"\bNone\b", text) is None, \
                "Found 'None' in yield result: %s" % text
            parts.append(text)
        ret = "\n".join(parts)
        if self.ENSURE_END_NEWLINE and not ret.endswith("\n"):
            ret += "\n"
        return ret



from ipaddress import IPv6Network

from annet.rpl_generators import AsPathFilter, CommunityList, CommunityType, RDFilter, ip_prefix_list, CommunityLogic, \
    IpPrefixListMember

AS_PATH_FILTERS = [
    AsPathFilter("ASP_EXAMPLE", [".*123456.*", ".*22.*"]),
]
COMMUNITIES = [
    CommunityList("COMMUNITY_EXAMPLE_ADD", ["1234:1000", "1234:1001"], logic=CommunityLogic.AND),
    CommunityList("COMMUNITY_EXAMPLE_REMOVE", ["1234:999", "1234:998"], logic=CommunityLogic.OR),
    CommunityList("EXTCOMMUNITY_EXAMPLE_ADD", ["12345:1000"], CommunityType.RT),
    CommunityList("EXTCOMMUNITY_EXAMPLE_REMOVE", ["12345:999"], CommunityType.RT),
]

RD_FILTERS = [
    RDFilter("RD_EXAMPLE1", 1, ["100:1", "200:1"]),
    RDFilter("RD_EXAMPLE2", 2, ["10.2.2.2:1", "10.3.3.3:1"]),
]

PREFIX_LISTS = [
    ip_prefix_list("IPV6_LIST_EXAMPLE", [
        "2a13:5941::/32",
        IpPrefixListMember(IPv6Network("2a13:5942::/32"), or_longer=(32, 48)),
        IpPrefixListMember(IPv6Network("2a13:5943::/32"), or_longer=(32, None)),
        IpPrefixListMember(IPv6Network("2a13:5944::/32"), or_longer=(32, 32)),
    ]),
    ip_prefix_list("IPV4_LIST_EXAMPLE", ["0.0.0.0/8", "10.0.0.0/8"]),
]



from annet.bgp_models import Redistribute, BFDTimers
from annet.mesh import Right, MeshRulesRegistry, GlobalOptions, MeshSession, DirectPeer, VirtualLocal, VirtualPeer

registry = MeshRulesRegistry()


@registry.device("{name:.*}")
def device_handler(global_opts: GlobalOptions):
    global_opts.local_as = 12345
    global_opts.ipv4_unicast.redistributes = (Redistribute(
        protocol="ipv4", policy="sss",
    ),)
    global_opts.groups["GROP_NAME"].remote_as = 11111


@registry.direct("{name:.*}", "m9-sgw{x}.{domain:.*}")
def direct_handler(device: DirectPeer, neighbor: DirectPeer, session: MeshSession):
    session.asnum = 12345
    device.addr = "192.168.1.254"
    neighbor.addr = f"192.168.1.{neighbor.match.x}"


@registry.virtual("{name:.*}", num=[1, 2, 3])
def virtual_handler(device: VirtualLocal, peer: VirtualPeer, session: MeshSession):
    session.asnum = 12345
    device.svi = 1
    device.addr = "192.168.1.254"
    device.listen_network = ["10.0.0.0/8"]
    peer.addr = f"192.168.127.{peer.num}"


@registry.direct("{name:.*}", "m9-sgw{x}.{domain:.*}", Right.x.in_([0, 1]))
def direct_handler2(device: DirectPeer, neighbor: DirectPeer, session: MeshSession):
    session.asnum = 12345
    device.addr = "192.168.1.254/24"
    device.lag = 1
    device.lag_links_min = neighbor.match.x
    device.subif = 100
    neighbor.families = {"ipv4_unicast"}
    neighbor.group_name = "GROUP_NAME"
    neighbor.addr = "192.168.1.200/24"



import os

from annet.annlib.command import Command, CommandList
from annet.annlib.netdev.views.hardware import HardwareView
from annet.vendors.tabparser import CommonFormatter
from annet.vendors.base import AbstractVendor
from annet.vendors.registry import registry


@registry.register
class PCVendor(AbstractVendor):
    NAME = "pc"

    def match(self) -> list[str]:
        return ["PC"]

    @property
    def reverse(self) -> str:
        return "-"

    def apply(self, hw: HardwareView, do_commit: bool, do_finalize: bool, path: str) -> tuple[CommandList, CommandList]:
        before, after = CommandList(), CommandList()

        if hw.soft.startswith(("Cumulus", "SwitchDev")):
            if os.environ.get("ETCKEEPER_CHECK", False):
                before.add_cmd(Command("etckeeper check"))

        return before, after

    @property
    def hardware(self) -> HardwareView:
        return HardwareView("PC")

    def make_formatter(self, **kwargs) -> CommonFormatter:
        return CommonFormatter(**kwargs)

    @property
    def exit(self) -> str:
        return ""



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any

from annet.generators import PartialGenerator
from annet.rpl import RouteMap, MatchField, RoutingPolicy
from .entities import AsPathFilter


def get_used_as_path_filters(
        as_path_filters: Sequence[AsPathFilter], policies: list[RoutingPolicy],
) -> Sequence[AsPathFilter]:
    filters = {c.name: c for c in as_path_filters}

    used_filters = set()
    for policy in policies:
        for statement in policy.statements:
            for condition in statement.match.find_all(MatchField.as_path_filter):
                used_filters.add(condition.value)
    return [filters[name] for name in sorted(used_filters)]


class AsPathFilterGenerator(PartialGenerator, ABC):
    TAGS = ["policy", "rpl", "routing"]

    @abstractmethod
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        raise NotImplementedError()

    @abstractmethod
    def get_as_path_filters(self, device: Any) -> Sequence[AsPathFilter]:
        raise NotImplementedError

    def get_used_as_path_filters(self, device: Any) -> Sequence[AsPathFilter]:
        filters = self.get_as_path_filters(device)
        policies = self.get_policies(device)
        return get_used_as_path_filters(filters, policies)

    def acl_huawei(self, _):
        return r"""
        ip as-path-filter
        """

    def run_huawei(self, device: Any):
        for as_path_filter in self.get_used_as_path_filters(device):
            values = "_".join((x for x in as_path_filter.filters if x != ".*"))
            yield "ip as-path-filter", as_path_filter.name, "index 10 permit", f"_{values}_"

    def acl_arista(self, _):
        return r"""
        ip as-path access-list
        """

    def run_arista(self, device: Any):
        for as_path_filter in self.get_used_as_path_filters(device):
            values = "_".join((x for x in as_path_filter.filters if x != ".*"))
            yield "ip as-path access-list", as_path_filter.name, "permit", f"_{values}_"

    def acl_iosxr(self, _):
        return r"""
        as-path-set *
            ~ %global=1
        """

    def run_iosxr(self, device: Any):
        for as_path_filter in self.get_used_as_path_filters(device):
            with self.block("as-path-set", as_path_filter.name):
                for n, filter_item in enumerate(as_path_filter.filters):
                    if n + 1 < len(as_path_filter.filters):
                        comma = ","
                    else:
                        comma = ""
                    yield "ios-regex", f"'{filter_item}'{comma}"

    def acl_juniper(self, _):
        return r"""
        policy-options  %cant_delete
            as-path ~
        """

    def _juniper_as_path(self, name: str, as_path_member: str):
        if not as_path_member.isnumeric():
            as_path_member = f'"{as_path_member}"'
        yield "as-path", name, as_path_member

    def run_juniper(self, device: Any):
        for as_path_filter in self.get_used_as_path_filters(device):
            # TODO could be implemented via as-path-groups
            # But we need to provide as_path_filters to policy generator
            # To select between regular as-path and as-path-groups
            if len(as_path_filter.filters) > 1:
                raise NotImplementedError(f"Multiple elements in as_path_filter {as_path_filter.name} is not supported for Juniper")

            with self.block("policy-options"):
                yield from self._juniper_as_path(as_path_filter.name, as_path_filter.filters[0])



from annet.annlib.types import Op

from annet.rulebook import common


def diff(old, new, diff_pre, _pops=(Op.AFFECTED,)):
    for iface_row in old:
        _filter_channel_members(old[iface_row])
    for iface_row in new:
        _filter_channel_members(new[iface_row])

    ret = common.default_diff(old, new, diff_pre, _pops)
    vpn_changed = False
    for (op, cmd, _, _) in ret:
        if op in {Op.ADDED, Op.REMOVED}:
            vpn_changed |= is_vpn_cmd(cmd)
    if vpn_changed:
        for cmd in list(old.keys()):
            if is_ip_cmd(cmd) and not is_vpn_cmd(cmd):
                del old[cmd]
        ret = common.default_diff(old, new, diff_pre, _pops)
    return ret


def is_vpn_cmd(cmd):
    return cmd.startswith(("ip vrf forwarding", "vrf forwarding"))


def is_ip_cmd(cmd):
    return cmd.startswith(("ip ", "ipv6 "))

# ===

# Вырезает все команды не разрешенные
# на членах агрегата. В running-config
# листинге они наследуются от самого port-channel


def _filter_channel_members(tree):
    if any(is_in_channel(x) for x in tree):
        for cmd in list(tree.keys()):
            if not _is_allowed_on_channel(cmd):
                del tree[cmd]


def is_in_channel(cmd_line):
    """
    Признак того, что это lagg member
    """
    return cmd_line.startswith("channel-group")


# Возможно тут есть еще какие-то команды
def _is_allowed_on_channel(cmd_line):
    return cmd_line.startswith((
        "channel-group",
        "cdp",
        "description",
        "inherit",
        "ip port",
        "ipv6 port",
        "mac port",
        "lacp",
        "switchport host",
        "shutdown",
        "rate-limit cpu",
        "snmp trap link-status",
    ))



from collections.abc import Sequence, Iterable
from dataclasses import dataclass, field
from typing import Literal, Union, Optional, Annotated


class VidRange:
    def __init__(self, start: int, stop: int) -> None:
        self.start = start
        self.stop = stop

    def is_single(self):
        return self.start == self.stop

    def __iter__(self):
        return iter(range(self.start, self.stop + 1))

    def __str__(self):
        if self.is_single():
            return str(self.start)
        return f"{self.start}-{self.stop}"

    def __repr__(self):
        return f"VlanRange({self.start}, {self.stop})"

    def __eq__(self, other: object) -> bool:
        if type(other) is VidRange:
            return self.start == other.start and self.stop == other.stop
        return NotImplemented


def _parse_vlan_ranges(ranges: str) -> Iterable[VidRange]:
    for range in ranges.split(","):
        start, sep, stop = range.strip().partition("-")
        try:
            if not sep:
                int_start = int(start)
                yield VidRange(int_start, int_start)
            elif not stop or not start:
                raise ValueError(f"Cannot parse range {range!r}. Expected `start-stop`")
            else:
                yield VidRange(int(start), int(stop))
        except ValueError:
            raise ValueError(f"Cannot parse range {range!r}. Expected `vid1-vid2` or `vid`")


class VidCollection:
    @staticmethod
    def parse(ranges: int | str) -> "VidCollection":
        if isinstance(ranges, int):
            return VidCollection([VidRange(ranges, ranges)])
        elif isinstance(ranges, str):
            return VidCollection(list(_parse_vlan_ranges(ranges)))
        elif isinstance(ranges, VidCollection):
            return VidCollection(ranges.ranges)
        else:
            raise TypeError(f"Expected str or int, got {type(ranges)}")

    def __init__(self, ranges: list[VidRange]) -> None:
        self.ranges = ranges

    def __str__(self):
        return ",".join(map(str, self.ranges))

    def __repr__(self):
        return f"VlanCollection({str(self)!r})"

    def __iter__(self):
        for range in self.ranges:
            yield from range

    def __eq__(self, other: object) -> bool:
        if type(other) is VidCollection:
            return self.ranges == other.ranges
        return False


class ASN(int):
    """
    Stores ASN number and formats it as в AS1.AS2
    None is treated like 0. Supports integer operations
    Supported formats: https://tools.ietf.org/html/rfc5396#section-1
    """
    PLAIN_MAX = 0x10000

    def __new__(cls, asn: Union[int, str, None, "ASN"]):
        if isinstance(asn, ASN):
            return asn
        elif asn is None:
            asn = 0
        elif not isinstance(asn, int):
            if isinstance(asn, str) and "." in asn:
                high, low = [int(token) for token in asn.split(".")]
                if not (0 <= high < ASN.PLAIN_MAX and 0 <= low < ASN.PLAIN_MAX):
                    raise ValueError("Invalid ASN asn %r" % asn)
                asn = (high << 16) + low
            asn = int(asn)
        if not 0 <= asn <= 0xffffffff:
            raise ValueError("Invalid ASN asn %r" % asn)
        return int.__new__(cls, asn)

    def __add__(self, other: int):
        return ASN(super().__add__(other))

    def __sub__(self, other: int):
        return ASN(super().__sub__(other))

    def __mul__(self, other: int):
        return ASN(super().__mul__(other))

    def is_plain(self) -> bool:
        return self < ASN.PLAIN_MAX

    def asdot(self) -> str:
        if not self.is_plain():
            return "%d.%d" % (self // ASN.PLAIN_MAX, self % ASN.PLAIN_MAX)
        return "%d" % self

    def asplain(self) -> str:
        return "%d" % self

    def asdot_high(self) -> int:
        return self // ASN.PLAIN_MAX

    def asdot_low(self) -> int:
        return self % ASN.PLAIN_MAX

    __str__ = asdot

    def __repr__(self) -> str:
        srepr = str(self)
        if "." in srepr:
            srepr = repr(srepr)
        return f"{self.__class__.__name__}({srepr})"


@dataclass(frozen=True)
class BFDTimers:
    minimum_interval: int = 500
    multiplier: int = 4


Family = Literal[
    "ipv4_unicast",
    "ipv6_unicast",
    "ipv4_vpn_unicast",
    "ipv6_vpn_unicast",
    "ipv4_labeled_unicast",
    "ipv6_labeled_unicast",
    "l2vpn_evpn",
]


@dataclass(frozen=True)
class PeerOptions:
    """The same options as for group but any field is optional"""
    local_as: Optional[ASN] = None
    unnumbered: Optional[bool] = None
    rr_client: Optional[bool] = None
    next_hop_self: Optional[bool] = None
    extended_next_hop: Optional[bool] = None
    send_community: Optional[bool] = None
    send_lcommunity: Optional[bool] = None
    send_extcommunity: Optional[bool] = None
    send_labeled: Optional[bool] = None
    import_limit: Optional[int] = None
    import_limit_action: Optional[str] = None
    teardown_timeout: Optional[bool] = None
    redistribute: Optional[bool] = None
    passive: Optional[bool] = None
    mtu_discovery: Optional[bool] = None
    advertise_inactive: Optional[bool] = None
    advertise_bgp_static: Optional[bool] = None
    allowas_in: Optional[bool] = None
    auth_key: Optional[bool] = None
    add_path: Optional[bool] = None
    multipath: Optional[bool] = None
    multihop: Optional[bool] = None
    multihop_no_nexthop_change: Optional[bool] = None
    af_no_install: Optional[bool] = None
    bfd: Optional[bool] = None
    rib: Optional[bool] = None
    bfd_timers: Optional[BFDTimers] = None
    resolve_vpn: Optional[bool] = None
    af_rib_group: Optional[str] = None
    af_loops: Optional[int] = None
    hold_time: Optional[int] = None
    listen_network: Optional[list[str]] = None
    remove_private: Optional[bool] = None
    as_override: Optional[bool] = None
    aigp: Optional[bool] = None
    bmp_monitor: Optional[bool] = None
    no_prepend: Optional[bool] = None
    no_explicit_null: Optional[bool] = None
    uniq_iface: Optional[bool] = None
    advertise_peer_as: Optional[bool] = None
    connect_retry: Optional[bool] = None
    advertise_external: Optional[bool] = None
    advertise_irb: Optional[bool] = None
    listen_only: Optional[bool] = None
    soft_reconfiguration_inbound: Optional[bool] = None
    not_active: Optional[bool] = None
    mtu: Optional[int] = None
    password: Optional[str] = None
    cluster_id: Optional[str] = None


@dataclass
class PeerFamilyOption:
    af_loops: Optional[int] = None
    import_limit: Optional[int] = None


@dataclass
class PeerFamilyOptions:
    ipv4_unicast: PeerFamilyOption = field(default_factory=PeerFamilyOption)
    ipv6_unicast: PeerFamilyOption = field(default_factory=PeerFamilyOption)
    ipv4_vpn_unicast: PeerFamilyOption = field(default_factory=PeerFamilyOption)
    ipv6_vpn_unicast: PeerFamilyOption = field(default_factory=PeerFamilyOption)
    ipv4_labeled_unicast: PeerFamilyOption = field(default_factory=PeerFamilyOption)
    ipv6_labeled_unicast: PeerFamilyOption = field(default_factory=PeerFamilyOption)
    l2vpn_evpn: PeerFamilyOption = field(default_factory=PeerFamilyOption)


@dataclass
class Peer:
    addr: str
    interface: Optional[str]
    remote_as: ASN
    families: set[Family] = field(default_factory=set)
    family_options: PeerFamilyOptions = field(default_factory=PeerFamilyOptions)
    description: str = ""
    vrf_name: str = ""
    group_name: str = ""
    import_policy: str = ""
    export_policy: str = ""
    update_source: Optional[str] = None
    options: Optional[PeerOptions] = None
    hostname: str = ""


@dataclass
class Aggregate:
    policy: str = ""
    routes: tuple[str, ...] = ()  # "182.168.1.0/24",
    as_path: str = ""
    reference: str = ""
    suppress: bool = False
    discard: bool = True
    as_set: bool = False


@dataclass
class Redistribute:
    protocol: str
    policy: str = ""


@dataclass
class FamilyOptions:
    family: Family
    vrf_name: str = ""
    multipath: int = 0
    global_multipath: int = 0
    aggregate: Aggregate = field(default_factory=Aggregate)  # use `aggregates` instead
    aggregates: tuple[Aggregate, ...] = ()
    af_loops: Optional[int] = None
    redistributes: tuple[Redistribute, ...] = ()
    allow_default: bool = False
    aspath_relax: bool = False
    igp_ignore: bool = False
    next_hop_policy: bool = False
    rib_import_policy: bool = False
    advertise_l2vpn_evpn: bool = False
    rib_group: bool = False
    loops: int = 0
    advertise_bgp_static: bool = False
    import_policy: Optional[str] = None
    export_policy: Optional[str] = None


@dataclass(frozen=True)
class PeerGroup:
    name: str
    remote_as: ASN = ASN(None)
    families: set[Family] = field(default_factory=set)
    family_options: PeerFamilyOptions = field(default_factory=PeerFamilyOptions)
    internal_name: str = ""
    description: str = ""
    update_source: str = ""
    peer_filter: str = ""
    import_policy: str = ""
    export_policy: str = ""

    # more strict version of PeerOptions
    local_as: ASN = ASN(None)
    unnumbered: bool = False
    rr_client: bool = False
    next_hop_self: bool = False
    extended_next_hop: bool = False
    send_community: bool = False
    send_lcommunity: bool = False
    send_extcommunity: bool = False
    send_labeled: bool = False
    import_limit: int = False
    import_limit_action: Optional[str] = None
    teardown_timeout: bool = False
    redistribute: bool = False
    passive: bool = False
    mtu_discovery: bool = False
    advertise_inactive: bool = False
    advertise_bgp_static: bool = False
    allowas_in: bool = False
    auth_key: bool = False
    add_path: bool = False
    multipath: bool = False
    multihop: Optional[int] = None
    multihop_no_nexthop_change: bool = False
    af_no_install: bool = False
    bfd: bool = False
    rib: bool = False
    bfd_timers: Optional[BFDTimers] = None
    resolve_vpn: bool = False
    af_rib_group: Optional[str] = None
    af_loops: int = 0
    hold_time: int = 0
    listen_network: list[str] = field(default_factory=list)
    remove_private: bool = False
    as_override: bool = False
    aigp: bool = False
    bmp_monitor: bool = False
    no_prepend: bool = False
    no_explicit_null: bool = False
    uniq_iface: bool = False
    advertise_peer_as: bool = False
    connect_retry: bool = False
    advertise_external: bool = False
    advertise_irb: bool = False
    listen_only: bool = False
    soft_reconfiguration_inbound: bool = False
    not_active: bool = False
    mtu: int = 0
    password: Optional[str] = None
    cluster_id: Optional[str] = None


@dataclass
class L2VpnOptions:
    name: str
    vid: VidCollection
    l2vni: int  # VNI, possible values are 1 to 2**24-1
    route_distinguisher: str = ""  # like in VrfOptions
    rt_import: list[str] = field(default_factory=list)  # like in VrfOptions
    rt_export: list[str] = field(default_factory=list)  # like in VrfOptions
    advertise_host_routes: bool = True  # advertise IP+MAC routes into L3VNI


@dataclass
class VrfOptions:
    vrf_name: str

    ipv4_unicast: FamilyOptions
    ipv6_unicast: FamilyOptions
    ipv4_labeled_unicast: FamilyOptions
    ipv6_labeled_unicast: FamilyOptions
    l2vpn_evpn: FamilyOptions

    vrf_name_global: Optional[str] = None
    import_policy: str = ""
    export_policy: str = ""
    as_path_relax: bool = False
    l3vni: Optional[int] = None
    rt_import: list[str] = field(default_factory=list)
    rt_export: list[str] = field(default_factory=list)
    rt_import_v4: list[str] = field(default_factory=list)
    rt_export_v4: list[str] = field(default_factory=list)
    route_distinguisher: Optional[str] = None
    static_label: Optional[int] = None  # FIXME: str?

    groups: list[PeerGroup] = field(default_factory=list)


@dataclass
class GlobalOptions:
    ipv4_unicast: FamilyOptions
    ipv6_unicast: FamilyOptions
    ipv4_labeled_unicast: FamilyOptions
    ipv6_labeled_unicast: FamilyOptions
    l2vpn_evpn: FamilyOptions

    as_path_relax: bool = False
    local_as: ASN = ASN(None)
    loops: int = 0
    multipath: int = 0
    router_id: str = ""
    cluster_id: Optional[str] = None

    vrf: dict[str, VrfOptions] = field(default_factory=dict)
    groups: list[PeerGroup] = field(default_factory=list)
    l2vpn: dict[str, L2VpnOptions] = field(default_factory=dict)


@dataclass
class BgpConfig:
    global_options: GlobalOptions
    peers: list[Peer]


def _used_policies(peer: Union[Peer, PeerGroup, VrfOptions]) -> Iterable[str]:
    if peer.import_policy:
        yield peer.import_policy
    if peer.export_policy:
        yield peer.export_policy


def _used_families_policies(opts: Union[GlobalOptions, VrfOptions]) -> Iterable[str]:
    for family_opts in (
        opts.ipv4_unicast,
        opts.ipv6_unicast,
        opts.ipv4_labeled_unicast,
        opts.ipv6_labeled_unicast,
        opts.l2vpn_evpn,
    ):
        for red in family_opts.redistributes:
            if red.policy:
                yield red.policy
        if family_opts.aggregate and family_opts.aggregate.policy:
            yield family_opts.aggregate.policy
        for aggregate in family_opts.aggregates:
            if aggregate.policy:
                yield aggregate.policy
        if family_opts.export_policy:
            yield family_opts.export_policy
        if family_opts.import_policy:
            yield family_opts.import_policy


def extract_policies(config: BgpConfig) -> Sequence[str]:
    result: list[str] = []
    for vrf in config.global_options.vrf.values():
        for group in vrf.groups:
            result.extend(_used_policies(group))
        result.extend(_used_families_policies(vrf))
        result.extend(_used_policies(vrf))
    for group in config.global_options.groups:
        result.extend(_used_policies(group))
    for peer in config.peers:
        result.extend(_used_policies(peer))
    result.extend(_used_families_policies(config.global_options))
    return result



import ssl
from ipaddress import ip_interface
from logging import getLogger
from typing import Any, Optional, List, Union, Dict, cast, Generic, TypeVar

from annetbox.v37 import models as api_models

from annet.adapters.netbox.common.query import NetboxQuery, FIELD_VALUE_SEPARATOR
from annet.adapters.netbox.common.storage_opts import NetboxStorageOpts
from annet.storage import Storage
from .adapter import NetboxAdapter
from .models import (
    IpAddress, Interface, NetboxDevice, Prefix, FHRPGroup, FHRPGroupAssignment,
)

logger = getLogger(__name__)
NetboxDeviceT = TypeVar("NetboxDeviceT", bound=NetboxDevice)
InterfaceT = TypeVar("InterfaceT", bound=Interface)
IpAddressT = TypeVar("IpAddressT", bound=IpAddress)
PrefixT = TypeVar("PrefixT", bound=Prefix)
FHRPGroupT = TypeVar("FHRPGroupT", bound=FHRPGroup)
FHRPGroupAssignmentT = TypeVar(
    "FHRPGroupAssignmentT", bound=FHRPGroupAssignment,
)


class BaseNetboxStorage(
    Storage,
    Generic[
        NetboxDeviceT,
        InterfaceT,
        IpAddressT,
        PrefixT,
        FHRPGroupT,
        FHRPGroupAssignmentT,
    ],
):
    """
    Base class for Netbox storage
    """

    def __init__(self, opts: Optional[NetboxStorageOpts] = None):
        ctx: Optional[ssl.SSLContext] = None
        url = ""
        token = ""
        self.exact_host_filter = False
        threads = 1
        if opts:
            if opts.insecure:
                ctx = ssl.create_default_context()
                ctx.check_hostname = False
                ctx.verify_mode = ssl.CERT_NONE
            url = opts.url
            token = opts.token
            threads = opts.threads
            self.exact_host_filter = opts.exact_host_filter
            self.all_hosts_filter = opts.all_hosts_filter
        self.netbox = self._init_adapter(url=url, token=token, ssl_context=ctx, threads=threads)
        self._all_fqdns: Optional[list[str]] = None
        self._id_devices: dict[int, NetboxDeviceT] = {}
        self._name_devices: dict[str, NetboxDeviceT] = {}
        self._short_name_devices: dict[str, NetboxDeviceT] = {}

    def _init_adapter(
            self,
            url: str,
            token: str,
            ssl_context: Optional[ssl.SSLContext],
            threads: int,
    ) -> NetboxAdapter[NetboxDeviceT, InterfaceT, IpAddressT, PrefixT, FHRPGroupT, FHRPGroupAssignmentT]:
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, _, __, ___):
        pass

    def resolve_object_ids_by_query(self, query: NetboxQuery):
        return [
            d.id for d in self._load_devices(query)
        ]

    def resolve_fdnds_by_query(self, query: NetboxQuery):
        return [
            d.name for d in self._load_devices(query)
        ]

    def resolve_all_fdnds(self) -> list[str]:
        if self._all_fqdns is None:
            self._all_fqdns = self.netbox.list_fqdns(self.all_hosts_filter)
        return self._all_fqdns

    def make_devices(
            self,
            query: Union[NetboxQuery, list],
            preload_neighbors=False,
            use_mesh=None,
            preload_extra_fields=False,
            **kwargs,
    ) -> List[NetboxDeviceT]:
        if isinstance(query, list):
            query = NetboxQuery.new(query)

        devices = []
        if query.is_host_query():
            globs = []
            for glob in query.globs:
                if glob in self._name_devices:
                    devices.append(self._name_devices[glob])
                if glob in self._short_name_devices:
                    devices.append(self._short_name_devices[glob])
                else:
                    globs.append(glob)
            if not globs:
                return devices
            query = NetboxQuery.new(globs)

        new_devices = self._load_devices(query)
        self._fill_device_interfaces(new_devices)
        for device in new_devices:
            self._record_device(device)
        return devices + new_devices

    def _load_devices(self, query: NetboxQuery) -> List[NetboxDeviceT]:
        if not query.globs:
            return []
        query_groups = parse_glob(self.exact_host_filter, query)
        devices = [
            device
            for device in self.netbox.list_devices(query_groups)
            if _match_query(self.exact_host_filter, query, device)
        ]
        return devices

    def _fill_device_interfaces(self, devices: list[NetboxDeviceT]) -> None:
        device_mapping = {d.id: d for d in devices}
        interfaces = self.netbox.list_interfaces_by_devices(list(device_mapping))
        for interface in interfaces:
            device_mapping[interface.device.id].interfaces.append(interface)
        self._fill_interface_fhrp_groups(interfaces)
        self._fill_interface_ipaddress(interfaces)

    def _fill_interface_fhrp_groups(self, interfaces: list[InterfaceT]) -> None:
        interface_mapping = {i.id: i for i in interfaces if i.count_fhrp_groups}
        assignments = self.netbox.list_fhrp_group_assignments(list(interface_mapping))
        group_ids = {r.fhrp_group_id for r in assignments}
        groups = {
            g.id: g for g in self.netbox.list_fhrp_groups(list(group_ids))
        }
        for assignment in assignments:
            assignment.group = groups[assignment.fhrp_group_id]
            interface_mapping[assignment.interface_id].fhrp_groups.append(assignment)

    def _fill_interface_ipaddress(self, interfaces: list[InterfaceT]) -> None:
        interface_mapping = {i.id: i for i in interfaces if i.count_ipaddresses}
        ips = self.netbox.list_ipaddr_by_ifaces(list(interface_mapping))
        for ip in ips:
            interface_mapping[ip.assigned_object_id].ip_addresses.append(ip)
        self._fill_ipaddr_prefixes(ips)

    def _fill_ipaddr_prefixes(self, ips: list[IpAddressT]) -> None:
        ip_to_cidrs: Dict[str, str] = {ip.address: str(ip_interface(ip.address).network) for ip in ips}
        prefixes = self.netbox.list_ipprefixes(list(ip_to_cidrs.values()))
        cidr_to_prefix: Dict[str, PrefixT] = {x.prefix: x for x in prefixes}
        for ip in ips:
            cidr = ip_to_cidrs[ip.address]
            ip.prefix = cidr_to_prefix.get(cidr)

    def _record_device(self, device: NetboxDeviceT):
        self._id_devices[device.id] = device
        self._short_name_devices[device.name] = device
        if not self.exact_host_filter:
            short_name = device.name.split(".")[0]
            self._short_name_devices[short_name] = device

    def get_device(
            self, obj_id, preload_neighbors=False, use_mesh=None,
            **kwargs,
    ) -> NetboxDeviceT:
        if obj_id in self._id_devices:
            return self._id_devices[obj_id]

        device = self.netbox.get_device(obj_id)
        self._fill_device_interfaces([device])
        self._record_device(device)
        return device

    def flush_perf(self):
        pass

    def search_connections(
            self,
            device: NetboxDeviceT,
            neighbor: NetboxDeviceT,
    ) -> list[tuple[InterfaceT, InterfaceT]]:
        if device.storage is not self:
            raise ValueError("device does not belong to this storage")
        if neighbor.storage is not self:
            raise ValueError("neighbor does not belong to this storage")
        # both devices are NetboxDevice if they are loaded from this storage
        res = []
        for local_port in device.interfaces:
            if not local_port.connected_endpoints:
                continue
            for endpoint in local_port.connected_endpoints:
                if endpoint.device.id == neighbor.id:
                    for remote_port in neighbor.interfaces:
                        if remote_port.name == endpoint.name:
                            res.append((local_port, remote_port))
                            break
        return res


def _match_query(exact_host_filter: bool, query: NetboxQuery, device_data: api_models.Device) -> bool:
    """
    Additional filtering after netbox due to limited backend logic.
    """
    if exact_host_filter:
        return True  # nothing to check, all filtering is done by netbox
    hostnames = [subquery.strip() for subquery in query.globs if FIELD_VALUE_SEPARATOR not in subquery]
    if not hostnames:
        return True  # no hostnames to check

    short_name = device_data.name.split(".")[0]
    for hostname in hostnames:
        hostname = hostname.strip().rstrip(".")
        if short_name == hostname or device_data.name == hostname:
            return True
    return False


def _hostname_dot_hack(raw_query: str) -> str:
    # there is no proper way to lookup host by its hostname
    # ie find "host" with fqdn "host.example.com"
    # besides using name__ic (ie startswith)
    # since there is no direct analogue for this field in netbox
    # so we need to add a dot to hostnames (top-level fqdn part)
    # so we would not receive devices with a common name prefix
    def add_dot(raw_query: Any) -> Any:
        if isinstance(raw_query, str) and "." not in raw_query:
            raw_query = raw_query + "."
        return raw_query

    if isinstance(raw_query, list):
        for i, name in enumerate(raw_query):
            raw_query[i] = add_dot(name)
    elif isinstance(raw_query, str):
        raw_query = add_dot(raw_query)

    return raw_query


def parse_glob(exact_host_filter: bool, query: NetboxQuery) -> dict[str, list[str]]:
    query_groups = cast(dict[str, list[str]], query.parse_query())
    if names := query_groups.pop("name", None):
        if exact_host_filter:
            query_groups["name__ie"] = names
        else:
            query_groups["name__ic"] = [_hostname_dot_hack(name) for name in names]
    return query_groups



import re
from collections import OrderedDict
from typing import Any

from annet.annlib.types import Op

from annet.rulebook import common


def ssh_key(rule, key, diff, hw, **_):
    """
    При включении ssh надо еще сгенерировать ключ. По конфигу никак не понять есть ли ключ на свитче или нет.
    """
    if diff[Op.ADDED]:
        added = sorted([x["row"] for x in diff[Op.ADDED]])
        if added == ["ip ssh version 2"]:
            # Отсыпаем mpdaemon-у подсказок для дополнительной команды при наливке
            comment = rule["comment"]
            rule["comment"] = ["!!suppress_errors!!", "!!timeout=240!!"]
            if hw.Cisco.C2960:
                yield (False, "crypto key generate rsa modulus 2048", None)
            else:
                yield (False, "crypto key generate rsa general-keys modulus 2048", None)
            rule["comment"] = comment
    yield from common.default(rule, key, diff)


def no_ipv6_nd_suppress_ra(rule, key, diff, **_):
    """
    При конфигурации ipv6 nd на нексусах нужно добавлять
    no ipv6 nd suppress-ra
    иначе RA не будет включен.
    К сожалению данной команды не видно в running-config.
    Поэтому подмешиваем ее в патч вместо генератора
    """
    if diff[Op.ADDED]:
        yield (False, "no ipv6 nd suppress-ra", None)
    yield from common.default(rule, key, diff)


def no_ntp_distribute(rule, key, diff, **_):
    """
    Для того, чтобы удалить NTP из CFS, сначала нужно сбросить активные
    NTP сессии.
    """
    if diff[Op.REMOVED]:
        yield (False, "clear ntp session", None)
    yield from common.default(rule, key, diff)


def banner_any(rule, key, diff, **_):
    if diff[Op.ADDED]:
        # Убираем дополнительный экранирующий сиимвол
        banner = re.sub(r"\^C", "^", diff[Op.ADDED][0]["row"])
        yield (False, banner, None)
    elif diff[Op.REMOVED]:
        yield (False, f"no banner {key[0]}", None)
    else:
        yield from common.default(rule, key, diff)



from typing import List

from annet.generators import BaseGenerator, PartialGenerator
from annet.storage import Storage


class Hostname(PartialGenerator):
    TAGS = ["hostname"]

    def acl_huawei(self, _):
        return """
        sysname
        header
        """

    def acl_safe_huawei(self, _):
        return """
        sysname
        header
        """

    def run_huawei(self, device):
        yield "sysname", device.hostname
        yield 'header login information "%s"' % device.hostname

    def acl_cisco(self, _):
        return """
            hostname
            banner login
        """

    def acl_safe_cisco(self, _):
        return """
            hostname
            banner login
        """

    def run_cisco(self, device):
        if device.hw.Cisco.AIR and device.tags["Lightweight WiFi точка"]:
            return
        elif device.hw.Catalyst:
            yield "banner login ^C%s^C" % device.hostname
        yield "hostname", device.hostname

    def acl_arista(self, _):
        return """
            hostname
        """

    def acl_safe_arista(self, _):
        return """
            hostname
        """

    def run_arista(self, device):
        yield "hostname", device.hostname

    acl_nexus = acl_arista
    run_nexus = run_arista

    def acl_juniper(self, _):
        return """
            system  %cant_delete
                host-name
        """

    def acl_safe_juniper(self, _):
        return """
            system  %cant_delete
                host-name
        """

    def run_juniper(self, device):
        with self.block("system"):
            yield "host-name", device.hostname

    acl_ribbon = acl_juniper
    acl_safe_ribbon = acl_safe_juniper
    run_ribbon = run_juniper

    def acl_nokia(self, _):
        return """
            system
                name
        """

    def acl_safe_nokia(self, _):
        return """
            system
                name
        """

    def run_nokia(self, device):
        with self.block("system"):
            yield "name", self.literal(device.hostname)

    def acl_routeros(self, _):
        return """
        system              %cant_delete
            identity        %cant_delete
                set ~       %cant_delete
        """

    def run_routeros(self, device):
        with self.block("system"):
            with self.block("identity"):
                yield f"set name={device.hostname}"


def get_generators(store: Storage) -> List[BaseGenerator]:
    return [
        Hostname(store),
    ]



from typing import TypeVar

from annet.bgp_models import extract_policies
from annet.mesh import MeshExecutor
from annet.rpl import RouteMap, RoutingPolicy

DeviceT = TypeVar("DeviceT")


def get_policies(routemap: RouteMap[DeviceT], mesh_executor: MeshExecutor, device: DeviceT) -> list[RoutingPolicy]:
    allowed_policies = extract_policies(mesh_executor.execute_for(device))
    return routemap.apply(device, allowed_policies)



from annet.mesh import MeshRulesRegistry, GlobalOptions

registry = MeshRulesRegistry()


@registry.device("{name:.*}")
def device_handler(global_opts: GlobalOptions):
    global_opts.groups["GROUP1"].import_policy = "example1"
    global_opts.groups["GROUP1"].export_policy = "example2"



import dataclasses
from dataclasses import dataclass, fields
from typing import Any, Iterable, List, Optional, Sequence

import yaml

from annet.annlib.netdev.views.dump import DumpableView
from annet.annlib.netdev.views.hardware import HardwareView
from annet.connectors import AdapterWithName
from annet.hardware import hardware_connector
from annet.storage import Device as DeviceProtocol
from annet.storage import Query, Storage, StorageProvider


@dataclass
class Interface(DumpableView):
    name: str
    description: str
    enabled: bool = True

    @property
    def _dump__list_key(self):
        return self.name


@dataclass
class DeviceStorage:
    fqdn: str

    vendor: Optional[str] = None
    hw_model: Optional[str] = None
    sw_version: Optional[str] = None

    hostname: Optional[str] = None
    serial: Optional[str] = None
    id: Optional[str] = None
    interfaces: Optional[list[Interface]] = None
    storage: Optional[Storage] = None

    hw: HardwareView = dataclasses.field(init=False)

    def __post_init__(self):
        if not self.id:
            self.id = self.fqdn
        if not self.hostname:
            self.hostname = self.fqdn.split(".")[0]

        if self.hw_model:
            hw = HardwareView(self.hw_model, self.sw_version)
            if self.vendor and self.vendor != hw.vendor:
                raise Exception(f"Vendor {self.vendor} is not vendor from hw model ({hw.vendor})")
            else:
                self.vendor = hw.vendor
        else:
            hw_provider = hardware_connector.get()
            hw: HardwareView = hw_provider.vendor_to_hw(self.vendor)
            if not hw:
                raise Exception("unknown vendor")
            self.hw_model = hw.model
        self.hw = hw

        if isinstance(self.interfaces, list):
            interfaces = []
            for iface in self.interfaces:
                try:
                    interfaces.append(Interface(**iface))
                except Exception as e:
                    raise Exception("unable to parse %s as Interface: %s" % (iface, e))
            self.interfaces = interfaces

    def set_storage(self, storage: Storage):
        self.storage = storage


@dataclass
class Device(DeviceProtocol, DumpableView):
    dev: DeviceStorage

    def __hash__(self):
        return hash((self.id, type(self)))

    def __eq__(self, other):
        return type(self) is type(other) and self.fqdn == other.fqdn and self.vendor == other.vendor

    def is_pc(self) -> bool:
        return False

    @property
    def hostname(self) -> str:
        return self.dev.hostname

    @property
    def fqdn(self) -> str:
        return self.dev.fqdn

    @property
    def id(self):
        return self.dev.id

    @property
    def storage(self) -> Storage:
        return self.dev.storage

    @property
    def hw(self) -> HardwareView:
        return self.dev.hw

    @property
    def breed(self) -> str:
        return self.dev.hw.vendor

    @property
    def neighbours_ids(self):
        return []

    @property
    def neighbours_fqdns(self) -> list[str]:
        return []

    def make_lag(self, lag: int, ports: Sequence[str], lag_min_links: Optional[int]) -> Interface:
        raise NotImplementedError

    def add_svi(self, svi: int) -> Interface:
        raise NotImplementedError

    def add_subif(self, interface: str, subif: int) -> Interface:
        raise NotImplementedError

    def find_interface(self, name: str) -> Optional[Interface]:
        raise NotImplementedError

    def flush_perf(self):
        pass


@dataclass
class Devices:
    devices: list[Device]

    def __post_init__(self):
        if isinstance(self.devices, list):
            devices = []
            for dev in self.devices:
                try:
                    devices.append(Device(dev=DeviceStorage(**dev)))
                except Exception as e:
                    raise Exception("unable to parse %s as Device: %s" % (dev, e))
            self.devices = devices


class Provider(StorageProvider, AdapterWithName):
    def storage(self):
        return storage_factory

    def opts(self):
        return StorageOpts

    def query(self):
        return Query

    @classmethod
    def name(cls) -> str:
        return "file"


@dataclass
class Query(Query):
    query: List[str]

    @classmethod
    def new(cls, query: str | Iterable[str], hosts_range: Optional[slice] = None) -> "Query":
        if hosts_range is not None:
            raise ValueError("host_range is not supported")
        return cls(query=list(query))

    @property
    def globs(self):
        return self.query

    def is_empty(self) -> bool:
        return len(self.query) == 0


class StorageOpts:
    def __init__(self, path: str):
        self.path = path

    @classmethod
    def parse_params(cls, conf_params: Optional[dict[str, str]], cli_opts: Any):
        path = conf_params.get("path")
        if not path:
            raise Exception("empty path")
        return cls(path=path)


def storage_factory(opts: StorageOpts) -> Storage:
    return FS(opts)


class FS(Storage):
    def __init__(self, opts: StorageOpts):
        self.opts = opts
        self.inventory: Devices = read_inventory(opts.path, self)

    def __enter__(self):
        return self

    def __exit__(self, _, __, ___):
        pass

    def resolve_object_ids_by_query(self, query: Query) -> list[str]:
        result = filter_query(self.inventory.devices, query)
        return [dev.fqdn for dev in result]

    def resolve_fdnds_by_query(self, query: Query) -> list[str]:
        result = filter_query(self.inventory.devices, query)
        return [dev.fqdn for dev in result]

    def make_devices(
        self,
        query: Query | list,
        preload_neighbors=False,
        use_mesh=None,
        preload_extra_fields=False,
        **kwargs,
    ) -> list[Device]:
        if isinstance(query, list):
            query = Query.new(query)
        result = filter_query(self.inventory.devices, query)
        return result

    def get_device(self, obj_id: str, preload_neighbors=False, use_mesh=None, **kwargs) -> Device:
        result = filter_query(self.inventory.devices, Query.new(obj_id))
        if not result:
            raise Exception("not found")
        return result[0]

    def flush_perf(self):
        pass

    def resolve_all_fdnds(self) -> list[str]:
        return [d.fqdn for d in self.inventory.devices]

    def search_connections(self, device: "Device", neighbor: "Device") -> list[tuple["Interface", "Interface"]]:
        return []


def filter_query(devices: list[Device], query: Query) -> list[Device]:
    result: list[Device] = []
    for dev in devices:
        if dev.fqdn in query.query:
            result.append(dev)
    return result


def read_inventory(path: str, storage: Storage) -> Devices:
    with open(path, "r") as f:
        file_data = yaml.load(f, Loader=yaml.SafeLoader)
    res = dataclass_from_dict(Devices, file_data)
    for dev in res.devices:
        dev.dev.set_storage(storage)
    return res


def dataclass_from_dict(klass: type, d: dict[str, Any]):
    try:
        fieldtypes = {f.name: f.type for f in fields(klass)}
    except TypeError:
        return d
    return klass(**{f: dataclass_from_dict(fieldtypes[f], d[f]) for f in d})



from typing import Literal, Annotated, Union, Optional

from .basemodel import BaseMeshModel, Concat, Unite
from ..bgp_models import BFDTimers, PeerFamilyOptions

FamilyName = Literal["ipv4_unicast", "ipv6_unicast", "ipv4_labeled_unicast", "ipv6_labeled_unicast", "l2vpn_evpn"]


class _SharedOptionsDTO(BaseMeshModel):
    """
    Options which can be set on connected pair or group of peers
    """
    add_path: bool
    multipath: bool
    advertise_irb: bool
    send_labeled: bool
    send_community: bool
    bfd: bool
    bfd_timers: BFDTimers
    password: str


class MeshSession(_SharedOptionsDTO):
    """
    Options which are set on connected pair
    """
    asnum: Union[int, str]
    vrf: str
    families: Annotated[set[FamilyName], Unite()]
    group_name: str

    bmp_monitor: bool

    import_policy: str
    export_policy: str


class _OptionsDTO(_SharedOptionsDTO):
    """
    Options which can be set on group of peers or peer itself
    """
    def __init__(self, **kwargs):
        kwargs.setdefault("family_options", PeerFamilyOptions())
        super().__init__(**kwargs)
    unnumbered: bool
    rr_client: bool
    next_hop_self: bool
    extended_next_hop: bool
    send_lcommunity: bool
    send_extcommunity: bool
    import_limit: Optional[int]
    import_limit_action: Optional[str]
    teardown_timeout: bool
    redistribute: bool
    passive: bool
    mtu_discovery: bool
    advertise_inactive: bool
    advertise_bgp_static: bool
    allowas_in: bool
    auth_key: bool
    multihop: Optional[int]
    multihop_no_nexthop_change: bool
    af_no_install: bool
    rib: bool
    resolve_vpn: bool
    af_rib_group: Optional[str]
    af_loops: int
    hold_time: int
    listen_network: list[str]
    remove_private: bool
    as_override: bool
    aigp: bool
    no_prepend: bool
    no_explicit_null: bool
    uniq_iface: bool
    advertise_peer_as: bool
    connect_retry: bool
    advertise_external: bool
    listen_only: bool
    soft_reconfiguration_inbound: bool
    not_active: bool
    mtu: int
    family_options: PeerFamilyOptions
    cluster_id: Optional[str]


class DirectPeerDTO(MeshSession, _OptionsDTO):
    pod: int
    addr: str
    description: str
    update_source: str

    subif: int
    lag: Optional[int]
    lag_links_min: Optional[int]
    svi: Optional[int]


class IndirectPeerDTO(MeshSession, _OptionsDTO):
    pod: int
    addr: str
    description: str
    update_source: str

    ifname: Optional[str]
    subif: int
    svi: Optional[int]


class VirtualLocalDTO(MeshSession, _OptionsDTO):
    pod: int
    addr: str
    description: str
    update_source: str

    svi: int


class VirtualPeerDTO(MeshSession):
    addr: str
    description: str


class MeshPeerGroup(_OptionsDTO):
    name: str
    families: Annotated[set[FamilyName], Unite()]
    local_as: Union[int, str]
    remote_as: Union[int, str]
    internal_name: str
    update_source: str
    description: str
    peer_filter: str

    import_policy: str
    export_policy: str
